@@ -12,31 +12,11 @@
 
 #if	MSDOS
 #include <process.h>
-extern int getcurdrv __P_((VOID_A));
-# ifndef	_NOUSELFN
-extern char *preparefile __P_((char *, char *));
-# endif
 #else
 #include <pwd.h>
 #include <grp.h>
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int lastdrv;
-# if	MSDOS && !defined (_NOUSELFN)
-extern int checkdrive __P_((int));
-# endif
-#endif
-
-extern int mark;
-extern char fullpath[];
-extern char *tmpfilename;
-extern int physical_path;
-extern int noconv;
-#ifndef	_NODOSDRIVE
-extern int dosdrive;
-#endif
-
 #define	MAXTMPNAMLEN	8
 #define	LOCKEXT		"LCK"
 #ifdef	_NODOSDRIVE
@@ -54,7 +34,13 @@ extern int dosdrive;
 #define	O_ACCMODE	(O_RDONLY | O_WRONLY | O_RDWR)
 #endif
 #ifndef	ENOSPC
-#define	ENOSPC	EACCES
+#define	ENOSPC		EACCES
+#endif
+#ifndef	EIO
+#define	EIO		ENODEV
+#endif
+#ifndef	ETIMEDOUT
+#define	ETIMEDOUT	EIO
 #endif
 #ifndef	L_SET
 # ifdef	SEEK_SET
@@ -64,35 +50,55 @@ extern int dosdrive;
 # endif
 #endif	/* !L_SET */
 
+#if	MSDOS
+extern int getcurdrv __P_((VOID_A));
+#endif
+#if	MSDOS && !defined (_NOUSELFN)
+extern char *preparefile __P_((CONST char *, char *));
+#endif
+#if	MSDOS && !defined (_NOUSELFN) && !defined (_NODOSDRIVE)
+extern int checkdrive __P_((int));
+#endif
+
+extern int mark;
+extern char fullpath[];
+extern char *tmpfilename;
+extern int physical_path;
+extern int noconv;
+#ifndef	_NODOSDRIVE
+extern int dosdrive;
+#endif
+
 #ifndef	NOFLOCK
 static int NEAR fcntllock __P_((int, int));
 #endif
-static char *NEAR excllock __P_((char *, int));
+static char *NEAR excllock __P_((CONST char *, int));
 #ifdef	_NODOSDRIVE
 #define	nodoschdir	Xchdir
 #define	nodosgetwd	Xgetwd
 #define	nodosmkdir	Xmkdir
 #define	nodosrmdir	Xrmdir
 #else
-static int NEAR nodoschdir __P_((char *));
+static int NEAR nodoschdir __P_((CONST char *));
 static char *NEAR nodosgetwd __P_((char *));
-static int NEAR nodosmkdir __P_((char *, int));
-static int NEAR nodosrmdir __P_((char *));
+static int NEAR nodosmkdir __P_((CONST char *, int));
+static int NEAR nodosrmdir __P_((CONST char *));
 #endif
-static int NEAR cpfile __P_((char *, char *, struct stat *, struct stat *));
+static int NEAR cpfile __P_((CONST char *, CONST char *,
+		struct stat *, struct stat *));
 static VOID changemes __P_((VOID_A));
 static int NEAR genrand __P_((int));
-static int dounlink __P_((char *));
-static int dormdir __P_((char *));
+static int dounlink __P_((CONST char *));
+static int dormdir __P_((CONST char *));
 #ifndef	_NOWRITEFS
-static int NEAR isexist __P_((char *));
-static int NEAR realdirsiz __P_((char *, int, int, int, int));
+static int NEAR isexist __P_((CONST char *));
+static int NEAR realdirsiz __P_((CONST char *, int, int, int, int));
 static int NEAR getnamlen __P_((int, int, int, int, int));
-static int NEAR saferename __P_((char *, char *));
-static char *NEAR maketmpfile __P_((int, int, char *, char *));
+static int NEAR saferename __P_((CONST char *, CONST char *));
+static char *NEAR maketmpfile __P_((int, int, CONST char *, CONST char *));
 #if	!MSDOS
-static off_t NEAR getdirblocksize __P_((char *));
-static u_char *NEAR getentnum __P_((char *, off_t));
+static off_t NEAR getdirblocksize __P_((CONST char *));
+static u_char *NEAR getentnum __P_((CONST char *, off_t));
 #endif
 static VOID NEAR restorefile __P_((char *, char *, int));
 #endif	/* !_NOWRITEFS */
@@ -109,8 +115,9 @@ static int dosdrv = -1;
 
 
 #ifdef	_USEDOSEMU
-char *nodospath(path, file)
-char *path, *file;
+CONST char *nodospath(path, file)
+char *path;
+CONST char *file;
 {
 	if (!_dospath(file)) return(file);
 	path[0] = '.';
@@ -153,7 +160,7 @@ namelist *namep;
 	char path[MAXPATHLEN];
 #endif
 	struct stat st, lst;
-	char *cp;
+	CONST char *cp;
 
 	cp = nodospath(path, namep -> name);
 	if (Xlstat(cp, &lst) < 0 || stat2(cp, &st) < 0) return(-1);
@@ -283,7 +290,7 @@ CONST VOID_P vp1;
 CONST VOID_P vp2;
 {
 	treelist *tp1, *tp2;
-	char *cp1, *cp2;
+	CONST char *cp1, *cp2;
 	int tmp;
 
 	tp1 = (treelist *)vp1;
@@ -316,8 +323,8 @@ CONST VOID_P vp2;
 /*ARGSUSED*/
 struct dirent *searchdir(dirp, regexp, arcregstr)
 DIR *dirp;
-reg_t *regexp;
-char *arcregstr;
+CONST reg_t *regexp;
+CONST char *arcregstr;
 {
 	struct dirent *dp;
 #ifndef	_NOARCHIVE
@@ -350,7 +357,8 @@ int underhome(buf)
 char *buf;
 {
 	static char *homedir = NULL;
-	char *cp, cwd[MAXPATHLEN];
+	CONST char *cp;
+	char tmp[MAXPATHLEN], cwd[MAXPATHLEN];
 	int len;
 
 #if	MSDOS
@@ -365,9 +373,8 @@ char *buf;
 #ifndef	_NODOSDRIVE
 		if (dospath2(cp)) return(-1);
 #endif
-		if (_chdir2(cp) < 0) return(-1);
-		homedir = getwd2();
-		if (_chdir2(cwd) < 0) lostcwd(cwd);
+		if (getrealpath(cp, tmp, cwd) != tmp) return(-1);
+		homedir = strdup2(tmp);
 	}
 
 	if (buf && !physical_path) strcpy(cwd, fullpath);
@@ -389,11 +396,11 @@ char *buf;
 }
 
 int preparedir(dir)
-char *dir;
+CONST char *dir;
 {
 	struct stat st;
 	char tmp[MAXPATHLEN];
-	char *cp;
+	CONST char *cp;
 
 	cp = dir;
 #ifdef	_USEDOSPATH
@@ -435,50 +442,81 @@ int fd, mode;
 		F_RDLCK, F_WRLCK, F_UNLCK,
 # else	/* !USEFCNTLOCK */
 #  ifdef	USELOCKF
-		F_LOCK, F_LOCK, F_ULOCK,
+		F_TLOCK, F_TLOCK, F_ULOCK,
 #  else
-		LOCK_SH, LOCK_EX, LOCK_UN,
+		LOCK_SH | LOCK_NB, LOCK_EX | LOCK_NB, LOCK_UN,
 #  endif
 # endif	/* !USEFCNTLOCK */
 	};
 # ifdef	USEFCNTLOCK
 	struct flock lock;
 # endif
+	int i, n;
 
 # ifndef	_NODOSDRIVE
 	if (fd >= DOSFDOFFSET) return(0);
 # endif
 
+	n = -1;
+	errno = 0;
+	for (i = 0; i < LCK_MAXRETRY; i++) {
 # ifdef	USEFCNTLOCK
-	lock.l_type = lockmode[mode];
-	lock.l_start = lock.l_len = (off_t)0;
-	lock.l_whence = SEEK_SET;
-	return(fcntl(fd, F_SETLKW, &lock));
+		lock.l_type = lockmode[mode];
+		lock.l_start = lock.l_len = (off_t)0;
+		lock.l_whence = SEEK_SET;
+		n = fcntl(fd, F_SETLK, &lock);
 # else	/* !USEFCNTLOCK */
 #  ifdef	USELOCKF
-	return(lockf(fd, lockmode[mode], (off_t)0));
+		n = lockf(fd, lockmode[mode], (off_t)0);
 #  else
-	return(flock(fd, lockmode[mode]));
+		n = flock(fd, lockmode[mode]);
 #  endif
 # endif	/* !USEFCNTLOCK */
+		if (n >= 0) {
+			n = 1;
+			break;
+		}
+# ifdef	EACCES
+		else if (errno == EACCES) /*EMPTY*/;
+# endif
+# ifdef	EAGAIN
+		else if (errno == EAGAIN) /*EMPTY*/;
+# endif
+# ifdef	EWOULDBLOCK
+		else if (errno == EWOULDBLOCK) /*EMPTY*/;
+# endif
+		else break;
+
+		if (intrkey(-1)) {
+			errno = EINTR;
+			break;
+		}
+# if	!MSDOS || defined (DJGPP)
+		usleep(100000L);
+# endif
+	}
+	if (i >= LCK_MAXRETRY) return(0);
+
+	return(n);
 }
 #endif	/* !NOFLOCK */
 
 static char *NEAR excllock(file, mode)
-char *file;
+CONST char *file;
 int mode;
 {
 #if	MSDOS
 	char *ext;
 #endif
 	static char **locklist = NULL;
+	struct stat st;
 	char *cp, path[MAXPATHLEN];
 	int i, fd;
 
 	if (!file) {
 		if (locklist) {
 			for (i = 0; locklist[i]; i++) {
-				Xunlink(locklist[i]);
+				VOID_C unlink(locklist[i]);
 				free(locklist[i]);
 			}
 			free(locklist);
@@ -501,7 +539,6 @@ int mode;
 				locklist = NULL;
 			}
 		}
-		free(file);
 		return(NULL);
 	}
 
@@ -516,15 +553,32 @@ int mode;
 #else
 	snprintf2(path, sizeof(path), "%s.%s", file, LOCKEXT);
 #endif
-	for (;;) {
-		fd = Xopen(path, O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
-			0666 & ~tmpumask);
-		if (fd >= 0) break;
-		if (errno != EEXIST) return(NULL);
+
+	fd = -1;
+	for (i = 0; i < LCK_MAXRETRY; i++) {
+		if (mode != LCK_READ)
+			fd = Xopen(path,
+				O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
+				0666 & ~tmpumask);
+		else if (Xlstat(path, &st) < 0)
+			fd = (errno == ENOENT) ? 0 : -1;
+		else {
+			errno = EEXIST;
+			fd = -1;
+		}
+		if (fd >= 0 || errno != EEXIST) break;
+
+		if (intrkey(-1)) {
+			errno = EINTR;
+			break;
+		}
 #if	!MSDOS || defined (DJGPP)
 		usleep(100000L);
 #endif
 	}
+	if (i >= LCK_MAXRETRY) return((char *)nullstr);
+	if (fd < 0) return(NULL);
+	if (mode == LCK_READ) return((char *)file);
 
 	Xclose(fd);
 	cp = strdup2(path);
@@ -537,18 +591,24 @@ int mode;
 }
 
 lockbuf_t *lockopen(path, flags, mode)
-char *path;
+CONST char *path;
 int flags, mode;
 {
+#ifndef	NOFLOCK
+	int n;
+#endif
 	char *lckname;
 	lockbuf_t *lck;
-	int fd, lckflags, lckmode, duperrno;
+	int fd, err, lckflags, lckmode, duperrno;
 
+#ifdef	FAKEUNINIT
+	fd = -1;	/* fake for -Wuninitialized */
+#endif
 	lckname = NULL;
 	lckflags = 0;
 	lckmode = ((flags & O_ACCMODE) == O_RDONLY) ? LCK_READ : LCK_WRITE;
+	err = 0;
 
-	fd = -1;
 #ifndef	NOFLOCK
 	if (isnfs(path) <= 0) {
 		lckflags |= LCK_FLOCK;
@@ -557,7 +617,12 @@ int flags, mode;
 			if ((flags & O_ACCMODE) == O_WRONLY || errno != ENOENT)
 				return(NULL);
 		}
-		else if (fcntllock(fd, lckmode) < 0) lckflags &= ~LCK_FLOCK;
+		else if ((n = fcntllock(fd, lckmode)) <= 0) {
+			Xclose(fd);
+			if (n < 0) return(NULL);
+			err++;
+			lckflags &= ~LCK_FLOCK;
+		}
 		else if ((flags & O_TRUNC) && Xftruncate(fd, (off_t)0) < 0) {
 			duperrno = errno;
 			VOID_C fcntllock(fd, LCK_UNLOCK);
@@ -569,20 +634,31 @@ int flags, mode;
 #endif	/* !NOFLOCK */
 
 	if (!(lckflags & LCK_FLOCK)) {
-		if (fd >= 0) Xclose(fd);
-		if ((lckname = excllock(path, lckmode))) /*EMPTY*/;
-		else if ((flags & O_ACCMODE) != O_RDONLY) return(NULL);
+		if (!(lckname = excllock(path, lckmode))) return(NULL);
+		else if (lckmode == LCK_READ) {
+			if (lckname == (char *)nullstr) err++;
+			lckname = NULL;
+		}
+		else if (lckname == (char *)nullstr) return(NULL);
 
 		if ((fd = newdup(Xopen(path, flags, mode))) >= 0) /*EMPTY*/;
 		else if ((flags & O_ACCMODE) == O_WRONLY || errno != ENOENT) {
 			duperrno = errno;
-			VOID_C excllock(lckname, LCK_UNLOCK);
+			if (lckname) {
+				VOID_C excllock(lckname, LCK_UNLOCK);
+				free(lckname);
+			}
 			errno = duperrno;
 			return(NULL);
 		}
 	}
 
 	if (fd < 0) lckflags |= LCK_INVALID;
+	else if (err) {
+		fprintf2(stderr, "%k: %k\r\n", path, NOLCK_K);
+		if (isttyiomode) warning(0, HITKY_K);
+	}
+
 	lck = (lockbuf_t *)malloc2(sizeof(lockbuf_t));
 	lck -> fd = fd;
 	lck -> fp = NULL;
@@ -593,7 +669,7 @@ int flags, mode;
 }
 
 lockbuf_t *lockfopen(path, type, flags)
-char *path, *type;
+CONST char *path, *type;
 int flags;
 {
 	lockbuf_t *lck;
@@ -617,7 +693,10 @@ VOID lockclose(lck)
 lockbuf_t *lck;
 {
 	if (lck) {
-		if (lck -> name) VOID_C excllock(lck -> name, LCK_UNLOCK);
+		if (lck -> name) {
+			VOID_C excllock(lck -> name, LCK_UNLOCK);
+			free(lck -> name);
+		}
 		if (!(lck -> flags & LCK_INVALID)) {
 #ifndef	NOFLOCK
 			if (lck -> flags & LCK_FLOCK)
@@ -632,9 +711,14 @@ lockbuf_t *lck;
 }
 
 int touchfile(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
+#ifdef	USEUTIME
+	struct utimbuf times;
+#else
+	struct timeval tvp[2];
+#endif
 #ifndef	_NOEXTRAATTR
 	int i;
 #endif
@@ -647,20 +731,11 @@ struct stat *stp;
 	if (s_islnk(&st)) return(1);
 
 	ret = 0;
-#ifdef	FAKEUNINIT
-	duperrno = errno;	/* fake for -Wuninitialized */
-#endif
 #if	MSDOS
 	if (!(st.st_mode & S_IWRITE)) Xchmod(path, (st.st_mode | S_IWRITE));
 #endif
 
 	if (stp -> st_nlink & (TCH_ATIME | TCH_MTIME)) {
-#ifdef	USEUTIME
-		struct utimbuf times;
-#else
-		struct timeval tvp[2];
-#endif
-
 		if (!(stp -> st_nlink & TCH_ATIME))
 			stp -> st_atime = st.st_atime;
 		if (!(stp -> st_nlink & TCH_MTIME))
@@ -668,28 +743,25 @@ struct stat *stp;
 #ifdef	USEUTIME
 		times.actime = stp -> st_atime;
 		times.modtime = stp -> st_mtime;
-		if (Xutime(path, &times) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xutime(path, &times) >= 0) errno = duperrno;
+		else ret = -1;
 #else
 		tvp[0].tv_sec = stp -> st_atime;
 		tvp[0].tv_usec = 0;
 		tvp[1].tv_sec = stp -> st_mtime;
 		tvp[1].tv_usec = 0;
-		if (Xutimes(path, tvp) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xutimes(path, tvp) >= 0) errno = duperrno;
+		else ret = -1;
 #endif
 	}
 
 #ifdef	HAVEFLAGS
 	if (stp -> st_nlink & TCH_FLAGS) {
-		if (Xchflags(path, stp -> st_flags) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xchflags(path, stp -> st_flags) >= 0) errno = duperrno;
+		else ret = -1;
 	}
 #endif
 
@@ -712,36 +784,40 @@ struct stat *stp;
 			mode |= (st.st_mode & (stp -> st_size));
 		}
 #endif	/* !_NOEXTRAATTR */
-		if (Xchmod(path, mode) < 0) {
-			duperrno = errno;
-			ret = -1;
-		}
+		duperrno = errno;
+		if (Xchmod(path, mode) >= 0) errno = duperrno;
+		else ret = -1;
 	}
 #if	MSDOS
-	else if (!(st.st_mode & S_IWRITE)) Xchmod(path, st.st_mode);
+	else if (!(st.st_mode & S_IWRITE)) {
+		duperrno = errno;
+		Xchmod(path, st.st_mode);
+		errno = duperrno;
+	}
 #endif
 
 #ifndef	NOUID
 	if (stp -> st_nlink & (TCH_UID | TCH_GID)) {
 		if (!(stp -> st_nlink & TCH_UID)) stp -> st_uid = (uid_t)-1;
 		if (!(stp -> st_nlink & TCH_GID)) stp -> st_gid = (gid_t)-1;
-		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0) /*EMPTY*/;
-		else if (!(stp -> st_nlink & TCH_CHANGE))
+		duperrno = errno;
+		if (Xchown(path, stp -> st_uid, stp -> st_gid) >= 0)
+			errno = duperrno;
+		else if (!(stp -> st_nlink & TCH_CHANGE)) {
 			Xchown(path, (uid_t)-1, stp -> st_gid);
-		else {
-			duperrno = errno;
-			ret = -1;
+			errno = duperrno;
 		}
+		else ret = -1;
 	}
 #endif	/* !NOUID */
-	if (ret < 0) errno = duperrno;
+	if (stp -> st_nlink & TCH_IGNOREERR) ret = 0;
 
 	return(ret);
 }
 
 #ifndef	_NODOSDRIVE
 static int NEAR nodoschdir(path)
-char *path;
+CONST char *path;
 {
 	int n, dupdosdrive;
 
@@ -768,7 +844,7 @@ char *path;
 }
 
 int nodoslstat(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 	int n, dupdosdrive;
@@ -782,7 +858,7 @@ struct stat *stp;
 }
 
 static int NEAR nodosmkdir(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
 	int n, dupdosdrive;
@@ -796,7 +872,7 @@ int mode;
 }
 
 static int NEAR nodosrmdir(path)
-char *path;
+CONST char *path;
 {
 	int n, dupdosdrive;
 
@@ -812,7 +888,8 @@ char *path;
 VOID lostcwd(path)
 char *path;
 {
-	char *cp, buf[MAXPATHLEN];
+	CONST char *cp;
+	char buf[MAXPATHLEN];
 	int duperrno;
 
 	duperrno = errno;
@@ -835,19 +912,20 @@ char *path;
 
 #ifndef	NODIRLOOP
 int issamebody(src, dest)
-char *src, *dest;
+CONST char *src, *dest;
 {
 	struct stat st1, st2;
 
 	if (Xstat(src, &st1) < 0 || Xstat(dest, &st2) < 0) return(0);
+	if (st1.st_dev != st2.st_dev || st1.st_ino != st2.st_ino) return(0);
 
-	return(st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino);
+	return(1);
 }
 #endif	/* !NODIRLOOP */
 
 #ifndef	NOSYMLINK
 int cpsymlink(src, dest)
-char *src, *dest;
+CONST char *src, *dest;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
@@ -869,14 +947,18 @@ char *src, *dest;
 
 /*ARGSUSED*/
 static int NEAR cpfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
 #if	MSDOS
 	struct stat st;
+#endif
+#ifndef	NODIRLOOP
+	int issame;
 #endif
 	char buf[BUFSIZ];
-	int i, fd1, fd2, flags, mode, duperrno;
+	off_t total;
+	int n, fd1, fd2, tty, flags, mode, duperrno;
 
 #if	MSDOS
 	if (!stp2 && Xlstat(dest, &st) >= 0) stp2 = &st;
@@ -889,6 +971,9 @@ struct stat *stp1, *stp2;
 
 	flags = (O_BINARY | O_RDONLY);
 	mode = stp1 -> st_mode;
+#ifndef	NODIRLOOP
+	issame = issamebody(src, dest);
+#endif
 	if ((fd1 = Xopen(src, flags, mode)) < 0) return(-1);
 
 	flags = (O_BINARY | O_WRONLY | O_CREAT | O_TRUNC);
@@ -896,7 +981,7 @@ struct stat *stp1, *stp2;
 	mode |= S_IWRITE;
 #endif
 #ifndef	NODIRLOOP
-	if (issamebody(src, dest)) {
+	if (issame) {
 		flags |= O_EXCL;
 		if (Xunlink(dest) < 0) {
 			VOID_C Xclose(fd1);
@@ -909,38 +994,46 @@ struct stat *stp1, *stp2;
 		return(-1);
 	}
 
-#ifdef	FAKEUNINIT
-	duperrno = errno;	/* fake for -Wuninitialized */
-#endif
+	tty = isatty(fd1);
+	total = (off_t)0;
 	for (;;) {
-		if ((i = sureread(fd1, buf, BUFSIZ)) <= 0) {
-			duperrno = errno;
+		if (!tty && total >= stp1 -> st_size) n = 0;
+		else n = sureread(fd1, buf, sizeof(buf));
+		if (n < 0) break;
+		total += (off_t)n;
+		if (!n) {
+			if (!tty && total < stp1 -> st_size) {
+				duperrno = ETIMEDOUT;
+				n = -1;
+			}
 			break;
 		}
-		if (surewrite(fd2, buf, i) < 0) {
-			i = -1;
-			duperrno = errno;
-			break;
-		}
-		if (i < BUFSIZ) break;
+#ifndef	_NOEXTRACOPY
+		showprogress(n);
+#endif
+		if ((n = surewrite(fd2, buf, n)) < 0) break;
 	}
 
-	VOID_C Xclose(fd2);
-	VOID_C Xclose(fd1);
+	safeclose(fd2);
+	safeclose(fd1);
+#ifndef	_NOEXTRACOPY
+	fshowprogress(dest);
+#endif
 
-	if (i < 0) {
+	if (n < 0) {
+		duperrno = errno;
 		VOID_C Xunlink(dest);
 		errno = duperrno;
 		return(-1);
 	}
-	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME);
-#ifdef	_USEDOSCOPY
-	if (touchfile(dest, stp1) < 0) return(-1);
-#else
-# ifndef	_NOEXTRACOPY
-	if (inheritcopy && touchfile(dest, stp1) < 0) return(-1);
+	stp1 -> st_nlink = (TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
+#if	defined (_USEDOSCOPY) || !defined (_NOEXTRACOPY)
+# ifndef	_USEDOSCOPY
+	if (!inheritcopy) /*EMPTY*/;
+	else
 # endif
-#endif
+	if (touchfile(dest, stp1) < 0) return(-1);
+#endif	/* _USEDOSCOPY || !_NOEXTRACOPY */
 
 	return(0);
 }
@@ -951,7 +1044,7 @@ static VOID changemes(VOID_A)
 }
 
 int safecpfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
 #ifndef	_NODOSDRIVE
@@ -981,16 +1074,21 @@ struct stat *stp1, *stp2;
 }
 
 int safemvfile(src, dest, stp1, stp2)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp1, *stp2;
 {
-	if (Xrename(src, dest) >= 0) return(0);
+	if (Xrename(src, dest) >= 0) {
+#ifndef	_NOEXTRACOPY
+		fshowprogress(dest);
+#endif
+		return(0);
+	}
 	if (errno != EXDEV || s_isdir(stp1)) return(-1);
 	if (safecpfile(src, dest, stp1, stp2) < 0 || Xunlink(src) < 0)
 		return(-1);
 
 	stp1 -> st_nlink = (TCH_MODE | TCH_UID | TCH_GID
-		| TCH_ATIME | TCH_MTIME);
+		| TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	return (touchfile(dest, stp1));
 }
 
@@ -1071,8 +1169,7 @@ char *dir;
 #endif
 	deftmpdir = strdup2(path);
 #if	MSDOS
-	n = getcurdrv();
-	*path = (isupper2(n)) ? toupper2(*path) : tolower2(*path);
+	*path = (isupper2(getcurdrv())) ? toupper2(*path) : tolower2(*path);
 #endif
 
 	mask = 0777 & ~tmpumask;
@@ -1131,7 +1228,7 @@ char *dir;
 }
 
 int rmtmpdir(dir)
-char *dir;
+CONST char *dir;
 {
 	char path[MAXPATHLEN];
 
@@ -1152,7 +1249,7 @@ char *dir;
 }
 
 int opentmpfile(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
 	char *cp;
@@ -1194,7 +1291,7 @@ char *file;
 }
 
 int rmtmpfile(file)
-char *file;
+CONST char *file;
 {
 	if ((Xunlink(file) < 0 && errno != ENOENT) || rmtmpdir(NULL) < 0)
 		return(-1);
@@ -1203,7 +1300,7 @@ char *file;
 }
 
 static int dounlink(path)
-char *path;
+CONST char *path;
 {
 	if (Xunlink(path) < 0) return(APL_ERROR);
 
@@ -1211,7 +1308,7 @@ char *path;
 }
 
 static int dormdir(path)
-char *path;
+CONST char *path;
 {
 	if (isdotdir(path)) return(APL_OK);
 	if (Xrmdir(path) < 0) return(APL_ERROR);
@@ -1220,7 +1317,8 @@ char *path;
 }
 
 VOID removetmp(dir, file)
-char *dir, *file;
+char *dir;
+CONST char *file;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
@@ -1235,21 +1333,11 @@ char *dir, *file;
 		NULL, dormdir, ORD_NOPREDIR, NULL);
 	else if (Xunlink(nodospath(path, file)) < 0) warning(-1, file);
 
-#ifndef	_NODOSDRIVE
-	if (dosdrv >= 0) {
-		if (lastdrv >= 0) {
-# if	MSDOS
-			if (_chdir2(deftmpdir) < 0);
-			else
-# endif
-			shutdrv(lastdrv);
-		}
-		lastdrv = dosdrv;
-		dosdrv = -1;
-	}
-#endif	/* !_NODOSDRIVE */
-
 	if (_chdir2(fullpath) < 0) lostcwd(fullpath);
+#ifndef	_NODOSDRIVE
+	shutdrv(dosdrv);
+	dosdrv = -1;
+#endif
 	if (dir) {
 		if (*dir && rmtmpdir(dir) < 0) warning(-1, dir);
 		free(dir);
@@ -1257,7 +1345,7 @@ char *dir, *file;
 }
 
 int forcecleandir(dir, file)
-char *dir, *file;
+CONST char *dir, *file;
 {
 #if	!MSDOS
 	p_id_t pid;
@@ -1300,14 +1388,15 @@ int drive;
 }
 
 int tmpdosdupl(dir, dirp, single)
-char *dir, **dirp;
+CONST char *dir;
+char **dirp;
 int single;
 {
-	struct stat st;
-	char *cp, *tmpdir, path[MAXPATHLEN];
-# if	!MSDOS
+# ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 # endif
+	struct stat st;
+	char *cp, *tmpdir, path[MAXPATHLEN];
 	int i, drive;
 
 	if (!(drive = dospath2(dir))) return(0);
@@ -1324,6 +1413,7 @@ int single;
 		strcpy(cp, filelist[filepos].name);
 		st.st_mode = filelist[filepos].st_mode;
 		st.st_atime = st.st_mtime = filelist[filepos].st_mtim;
+		st.st_size = filelist[filepos].st_size;
 		if (cpfile(fnodospath(tmp, filepos), path, &st, NULL) < 0) {
 			warning(-1, path);
 			removetmp(tmpdir, NULL);
@@ -1335,6 +1425,7 @@ int single;
 		strcpy(cp, filelist[i].name);
 		st.st_mode = filelist[i].st_mode;
 		st.st_atime = st.st_mtime = filelist[i].st_mtim;
+		st.st_size = filelist[i].st_size;
 		if (cpfile(fnodospath(tmp, i), path, &st, NULL) < 0) {
 			warning(-1, path);
 			removetmp(tmpdir, NULL);
@@ -1342,8 +1433,11 @@ int single;
 		}
 	}
 
-	dosdrv = lastdrv;
-	lastdrv = -1;
+	if (preparedrv(dospath(fullpath, NULL), &dosdrv) < 0) {
+		warning(-1, fullpath);
+		removetmp(tmpdir, NULL);
+		return(-1);
+	}
 	if (_chdir2(tmpdir) < 0) {
 		warning(-1, tmpdir);
 		removetmp(tmpdir, NULL);
@@ -1356,16 +1450,16 @@ int single;
 
 int tmpdosrestore(drive, file)
 int drive;
-char *file;
+CONST char *file;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
-# if	!MSDOS
+# ifdef	_USEDOSEMU
 	char tmp[MAXPATHLEN];
 # endif
 
 	strcpy(gendospath(path, drive, '\0'), file);
-# if	!MSDOS
+# ifdef	_USEDOSEMU
 	file = nodospath(tmp, file);
 # endif
 	waitmes();
@@ -1378,7 +1472,7 @@ char *file;
 
 #ifndef	_NOWRITEFS
 static int NEAR isexist(file)
-char *file;
+CONST char *file;
 {
 	struct stat st;
 
@@ -1388,7 +1482,7 @@ char *file;
 }
 
 static int NEAR realdirsiz(s, dos, boundary, dirsize, ofs)
-char *s;
+CONST char *s;
 int dos, boundary, dirsize, ofs;
 {
 	int i, len, lfn, dot;
@@ -1442,7 +1536,7 @@ int size, dos, boundary, dirsize, ofs;
 }
 
 static int NEAR saferename(from, to)
-char *from, *to;
+CONST char *from, *to;
 {
 #ifdef	_USEDOSEMU
 	char fpath[MAXPATHLEN], tpath[MAXPATHLEN];
@@ -1460,8 +1554,11 @@ char *from, *to;
 /*ARGSUSED*/
 static char *NEAR maketmpfile(len, dos, tmpdir, old)
 int len, dos;
-char *tmpdir, *old;
+CONST char *tmpdir, *old;
 {
+#ifndef	PATHNOCASE
+	int i;
+#endif
 	char *fname, path[MAXPATHLEN];
 	int l, fd;
 
@@ -1477,12 +1574,8 @@ char *tmpdir, *old;
 	for (;;) {
 		genrandname(fname, len);
 #ifndef	PATHNOCASE
-		if (dos) {
-			int i;
-
-			for (i = 0; fname[i]; i++)
-				fname[i] = toupper2(fname[i]);
-		}
+		if (dos) for (i = 0; fname[i]; i++)
+			fname[i] = toupper2(fname[i]);
 #endif
 		if (tmpdir) {
 			strcpy(&(path[l]), fname);
@@ -1523,7 +1616,7 @@ char *tmpdir, *old;
 
 #if	!MSDOS
 static off_t NEAR getdirblocksize(dir)
-char *dir;
+CONST char *dir;
 {
 	struct stat st;
 
@@ -1533,7 +1626,7 @@ char *dir;
 }
 
 static u_char *NEAR getentnum(dir, bsiz)
-char *dir;
+CONST char *dir;
 off_t bsiz;
 {
 	struct stat st;
@@ -1600,7 +1693,8 @@ int fs;
 	struct dirent *dp;
 	int dos, boundary, dirsize, namofs;
 	int i, top, size, len, fnamp, ent;
-	char *cp, *tmpdir, **fnamelist, path[MAXPATHLEN];
+	CONST char *cp;
+	char *tmp, *tmpdir, **fnamelist, path[MAXPATHLEN];
 
 	switch (fs) {
 #if	!MSDOS
@@ -1735,7 +1829,7 @@ int fs;
 	Xclosedir(dirp);
 
 	if (ent > 0) {
-		if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) {
+		if (!(tmp = maketmpfile(len, dos, tmpdir, tmpdir))) {
 			warning(-1, tmpdir);
 			restorefile(tmpdir, path, fnamp);
 			freevar(fnamelist);
@@ -1743,7 +1837,7 @@ int fs;
 			return;
 		}
 		free(tmpdir);
-		tmpdir = cp;
+		tmpdir = tmp;
 		fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	}
 
@@ -1819,10 +1913,11 @@ int fs;
 	}
 #endif
 
-	if (!(cp = maketmpfile(len, dos, tmpdir, tmpdir))) warning(-1, tmpdir);
+	if (!(tmp = maketmpfile(len, dos, tmpdir, tmpdir)))
+		warning(-1, tmpdir);
 	else {
 		free(tmpdir);
-		tmpdir = cp;
+		tmpdir = tmp;
 	}
 	fnamp = strcatdelim2(path, tmpdir, NULL) - path;
 	snprintf2(&(path[fnamp]), sizeof(path) - fnamp, fnamelist[top]);
