@@ -13,6 +13,10 @@
 #include "kanji.h"
 #include "termemu.h"
 
+#ifndef	_NOIME
+#include "roman.h"
+#endif
+
 #ifdef	_NOORIGSHELL
 #include "termio.h"
 #include "wait.h"
@@ -20,13 +24,17 @@
 #include "system.h"
 # ifndef	NOALIAS
 extern int addalias __P_((char *, char *));
-extern int deletealias __P_((char *));
+extern int deletealias __P_((CONST char *));
 # endif
 #endif
 
+#ifdef	_NOIME
 #define	MAXPTYMENU	4
+#else
+#define	MAXPTYMENU	5
+#endif
 
-extern functable funclist[];
+extern CONST functable funclist[];
 extern int internal_status;
 extern int fdmode;
 extern int fdflags;
@@ -40,6 +48,9 @@ extern p_id_t emupid;
 extern int emufd;
 extern int parentfd;
 extern char *ptytmpfile;
+#ifndef	_NOIME
+extern int ime_cont;
+#endif
 
 static int (*lastfunc)__P_((VOID_A)) = NULL;
 static u_long lockflags = 0;
@@ -54,6 +65,7 @@ static int NEAR recvcommand __P_((int, command_t **, syntaxtree *));
 static int NEAR recvstree __P_((int, syntaxtree **, syntaxtree *));
 #endif
 static VOID NEAR recvchild __P_((int));
+static VOID NEAR sendkey __P_((int));
 
 
 int waitstatus(pid, options, statusp)
@@ -158,7 +170,7 @@ int c;
 }
 
 VOID Xcputs2(s)
-char *s;
+CONST char *s;
 {
 	int len;
 
@@ -256,7 +268,7 @@ VOID Xcputnl(VOID_A)
 }
 
 int Xkanjiputs(s)
-char *s;
+CONST char *s;
 {
 	return(Xcprintf2("%k", s));
 }
@@ -384,12 +396,38 @@ VOID changeoutkcode(VOID_A)
 
 static int NEAR ptygetkey(VOID_A)
 {
-	char *cp, *str[MAXPTYMENU];
+#ifndef	_NOIME
+	static char buf[MAXKANJIBUF + 1] = "";
+	static int next = 0;
+	u_int w;
+#endif
+	CONST char *str[MAXPTYMENU];
+	char *cp, *new;
 	int n, c, ch, val[MAXPTYMENU];
 
+#ifndef	_NOIME
+	if (next > 0) {
+		if (next >= sizeof(buf) || !(c = (u_char)(buf[next++])))
+			next = 0;
+		else return(c);
+	}
+	if (ime_cont) {
+		c = ime_inputkanji(sigalrm(1), buf);
+		if (c < 0 || !ime_cont) movepos(filepos, 0);
+		if (c != K_ESC) {
+			if (c < 0) {
+				next = 0;
+				ime_inputkanji(0, NULL);
+			}
+			else if (!c && (c = (u_char)(buf[0]))) next = 1;
+			return(c);
+		}
+	}
+#endif	/* !_NOIME */
+
 	for (;;) {
 		n = -1;
-		c = getkey2(sigalrm(1));
+		c = getkey2(sigalrm(1), inputkcode);
 		while (lockflags & (1 << win)) {
 			kbhit2(1000000L / SENSEPERSEC);
 			waitpty();
@@ -397,16 +435,23 @@ static int NEAR ptygetkey(VOID_A)
 #ifndef	_NOORIGSHELL
 		if (isshptymode()) break;
 #endif
-		if (c < 0 || ptymenukey < 0 || c != ptymenukey) break;
+		if (c < 0 || ptymenukey < 0 || alternate(c) != ptymenukey)
+			break;
 
-		str[0] = asprintf3(PTYAI_K, getkeysym(ptymenukey, 0));
+		str[0] = new = asprintf3(PTYAI_K, getkeysym(ptymenukey, 0));
 		str[1] = PTYIC_K;
 		str[2] = PTYBR_K;
-		str[3] = PTYNW_K;
+#ifndef	_NOIME
+		str[3] = PTYKJ_K;
+#endif
+		str[MAXPTYMENU - 1] = PTYNW_K;
 		val[0] = 0;
 		val[1] = 1;
 		val[2] = 2;
+#ifndef	_NOIME
 		val[3] = 3;
+#endif
+		val[MAXPTYMENU - 1] = 4;
 
 		n = 0;
 		changewin(MAXWINDOWS, (p_id_t)-1);
@@ -414,7 +459,7 @@ static int NEAR ptygetkey(VOID_A)
 			0, str, val);
 		movepos(filepos, 0);
 		changewin(win, (p_id_t)-1);
-		free(str[0]);
+		free(new);
 
 		if (ch != K_CR) continue;
 		else if (!n) break;
@@ -424,11 +469,28 @@ static int NEAR ptygetkey(VOID_A)
 			c = -1;
 			break;
 		}
+#ifndef	_NOIME
 		else if (n == 3) {
+			c = ime_inputkanji(sigalrm(1), buf);
+			if (c < 0 || !ime_cont) movepos(filepos, 0);
+			if (c != K_ESC) {
+				if (c < 0) {
+					next = 0;
+					ime_inputkanji(0, NULL);
+				}
+				else if (!c && (c = (u_char)(buf[0])))
+					next = 1;
+				break;
+			}
+		}
+#endif	/* !_NOIME */
+#ifndef	_NOSPLITWIN
+		else if (n == 4) {
 			VOID_C nextwin();
 			c = -2;
 			break;
 		}
+#endif
 		else {
 			changewin(MAXWINDOWS, (p_id_t)-1);
 			lcmdline = -1;
@@ -437,9 +499,25 @@ static int NEAR ptygetkey(VOID_A)
 			movepos(filepos, 0);
 			changewin(win, (p_id_t)-1);
 			if (!cp) continue;
+#ifdef	_NOIME
 			c = getkeycode(cp, 0);
 			free(cp);
 			if (c >= 0) break;
+#else	/* !_NOIME */
+			w = ime_getkeycode(cp);
+			free(cp);
+			if (!w) /*EMPTY*/;
+			else if (!(w & ~01777)) {
+				c = w;
+				break;
+			}
+			else if ((c = ime_inkanjiconv(buf, w)) < 0) /*EMPTY*/;
+			else if (c || !(c = (u_char)(buf[0]))) break;
+			else {
+				next = 1;
+				break;
+			}
+#endif	/* !_NOIME */
 
 			changewin(MAXWINDOWS, (p_id_t)-1);
 			warning(0, VALNG_K);
@@ -448,6 +526,13 @@ static int NEAR ptygetkey(VOID_A)
 		}
 	}
 
+#ifndef	_NOIME
+	if (c < 0) {
+		next = 0;
+		ime_inputkanji(0, NULL);
+	}
+#endif
+
 	return(c);
 }
 
@@ -863,6 +948,20 @@ int w;
 			}
 			else if (cp) free(cp);
 			break;
+#ifndef	_NOIME
+		case TE_ADDROMAN:
+			if (recvstring(fd, &cp) < 0 || !cp) break;
+			if (recvstring(fd, &func1) < 0) {
+				free(cp);
+				break;
+			}
+			VOID_C addroman(cp, func1);
+			break;
+		case TE_FREEROMAN:
+			if (recvbuf(fd, &n, sizeof(n)) < 0) break;
+			freeroman(n);
+			break;
+#endif	/* !_NOIME */
 		case TE_INTERNAL:
 			if (recvbuf(fd, &n, sizeof(n)) < 0
 			|| recvstring(fd, &cp) < 0)
@@ -892,6 +991,58 @@ int w;
 	}
 }
 
+static VOID NEAR sendkey(c)
+int c;
+{
+	static char buf[MAXUTF8LEN * sizeof(short)];
+	static int index = 0;
+	static int max = 0;
+#ifndef	_NOKANJICONV
+	int code;
+#endif
+	short w;
+
+#ifndef	_NOKANJICONV
+	code = (inputkcode != NOCNV) ? inputkcode : DEFCODE;
+#endif
+	if (c >= 0) {
+		w = c;
+		memcpy(&(buf[index++ * sizeof(w)]), (char *)&w, sizeof(w));
+	}
+
+	if (c < 0 || c & 0xff00) /*EMPTY*/;
+	else if (max && index >= max) /*EMPTY*/;
+#ifdef	_NOKANJICONV
+# ifdef	CODEEUC
+	else if (c == C_EKANA) {
+		max = 2;
+		return;
+	}
+# endif
+#else	/* !_NOKANJICONV */
+	else if (code == EUC && c == C_EKANA) {
+		max = 2;
+		return;
+	}
+	else if (code >= UTF8) {
+		if (index > 1) {
+			if ((c & 0xc0) == 0x80) return;
+		}
+		else if (ismsb(c)) {
+			max = ((c & 0xf0) == 0xe0) ? 3 : 2;
+			return;
+		}
+	}
+#endif	/* !_NOKANJICONV */
+	else if (isinkanji1(c, code)) {
+		max = 2;
+		return;
+	}
+
+	if (index) sendbuf(emufd, buf, index * sizeof(w));
+	index = max = 0;
+}
+
 int frontend(VOID_A)
 {
 	int n, ch, wastty, status;
@@ -905,7 +1056,8 @@ int frontend(VOID_A)
 
 	internal_status = FNC_FAIL;
 	if (!(wastty = isttyiomode)) Xttyiomode(1);
-	while ((ch = ptygetkey()) >= 0) sendword(emufd, ch);
+	while ((ch = ptygetkey()) >= 0) sendkey(ch);
+	sendkey(-1);
 	if (!wastty) Xstdiomode();
 
 	if (ch < -1) {
