@@ -109,7 +109,7 @@ typedef struct _kconv_t {
 #if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
 typedef struct _langtable {
-	char *ident;
+	CONST char *ident;
 	u_char lang;
 } langtable;
 #endif
@@ -123,15 +123,15 @@ typedef struct _kpathtable {
 
 #if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
-static char *NEAR strstr2 __P_((char *, char *));
+static CONST char *NEAR strstr2 __P_((CONST char *, CONST char *));
 #endif
 #if	!defined (_NOKANJICONV) \
 || (defined (FD) && defined (_USEDOSEMU) && defined (CODEEUC))
-static VOID NEAR sj2j __P_((char *, u_char *));
-static VOID NEAR j2sj __P_((char *, u_char *));
+static VOID NEAR sj2j __P_((char *, CONST u_char *));
+static VOID NEAR j2sj __P_((char *, CONST u_char *));
 #endif
 #ifdef	_USEUNICODE
-static int NEAR openunitbl __P_((char *));
+static int NEAR openunitbl __P_((CONST char *));
 static u_char *NEAR newunitbl __P_((ALLOC_T));
 #define	getword(s, n)	(((u_short)((s)[(n) + 1]) << 8) | (s)[n])
 #define	skread(f,o,s,n)	(Xlseek(f, o, L_SET) >= (off_t)0 \
@@ -139,33 +139,34 @@ static u_char *NEAR newunitbl __P_((ALLOC_T));
 #endif
 #ifndef	_NOKANJICONV
 # ifdef	_USEUNICODE
-static int NEAR opennftbl __P_((char *, int, u_int *));
+static int NEAR opennftbl __P_((CONST char *, int, u_int *));
 # endif
-static int NEAR toenglish __P_((char *, u_char *, int));
-static int NEAR tojis7 __P_((char *, u_char *, int, int, int, int));
-static int NEAR fromjis __P_((char *, u_char *, int, int));
-static int NEAR tojis8 __P_((char *, u_char *, int, int, int, int));
-static int NEAR tojunet __P_((char *, u_char *, int, int, int, int));
+static int NEAR toenglish __P_((char *, CONST u_char *, int));
+static int NEAR tojis7 __P_((char *, CONST u_char *, int, int, int, int));
+static int NEAR fromjis __P_((char *, CONST u_char *, int, int));
+static int NEAR tojis8 __P_((char *, CONST u_char *, int, int, int, int));
+static int NEAR tojunet __P_((char *, CONST u_char *, int, int, int, int));
 # ifdef	_USEUNICODE
-static VOID NEAR ucs2normalization __P_((u_short *, int *, int, u_int, int));
-static u_int NEAR ucs2denormalization __P_((u_short *, int *, int));
-static u_int NEAR toucs2 __P_((u_char *, int *));
+static u_int NEAR toucs2 __P_((CONST u_char *, int *));
 static VOID NEAR fromucs2 __P_((char *, int *, u_int));
-static int NEAR toutf8 __P_((char *, u_char *, int));
-static int NEAR fromutf8 __P_((char *, u_char *, int));
-static int NEAR toutf8nf __P_((char *, u_char *, int, int));
-static int NEAR fromutf8nf __P_((char *, u_char *, int, int));
+static int NEAR toutf8 __P_((char *, CONST u_char *, int));
+static int NEAR fromutf8 __P_((char *, CONST u_char *, int));
+static int NEAR toutf8nf __P_((char *, CONST u_char *, int, int));
+static int NEAR fromutf8nf __P_((char *, CONST u_char *, int, int));
 # endif	/* _USEUNICODE */
 static int NEAR bin2hex __P_((char *, int));
-static int NEAR tohex __P_((char *, u_char *, int));
-static int NEAR fromhex __P_((char *, u_char *, int));
+static int NEAR tohex __P_((char *, CONST u_char *, int));
+static int NEAR fromhex __P_((char *, CONST u_char *, int));
 static int NEAR bin2cap __P_((char *, int));
-static int NEAR tocap __P_((char *, u_char *, int));
-static int NEAR fromcap __P_((char *, u_char *, int));
-static char *NEAR _kanjiconv __P_((char *, char *, int, int, int, int *, int));
+static int NEAR tocap __P_((char *, CONST u_char *, int));
+static int NEAR fromcap __P_((char *, CONST u_char *, int));
+static CONST char *NEAR _kanjiconv __P_((char *, CONST char *, int,
+		int, int, int *, int));
 #endif	/* !_NOKANJICONV */
 
+#ifdef	FD
 int noconv = 0;
+#endif
 #ifndef	_NOKANJIFCONV
 int nokanjifconv = 0;
 char *sjispath = NULL;
@@ -180,6 +181,7 @@ char *hexpath = NULL;
 char *cappath = NULL;
 char *utf8path = NULL;
 char *utf8macpath = NULL;
+char *utf8iconvpath = NULL;
 char *noconvpath = NULL;
 #endif	/* !_NOKANJIFCONV */
 #ifdef	_USEUNICODE
@@ -196,6 +198,7 @@ static CONST langtable langlist[] = {
 	{"ujis", EUC},
 	{"utf8-mac", M_UTF8},
 	{"mac", M_UTF8},
+	{"utf8-iconv", I_UTF8},
 	{"utf8", UTF8},
 	{"utf-8", UTF8},
 	{"ojunet", O_JUNET},
@@ -208,7 +211,8 @@ static CONST langtable langlist[] = {
 	{"cap", CAP},
 # endif	/* _NOKANJICONV */
 # ifndef	_NOENGMES
-	{"eng", ENG},
+	{"en", ENG},
+	{"POSIX", ENG},
 	{"C", ENG},
 # endif	/* _NOENGMES */
 };
@@ -328,7 +332,7 @@ static CONST u_char sj2jtable2[256] = {
 	0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,	/* 0xf0 */
 	0x7a, 0x7b, 0x7c, 0x7d, 0x7e,    0,    0,    0
 };
-static u_char j2sjtable1[128] = {
+static CONST u_char j2sjtable1[128] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
 	   0, 0x81, 0x81, 0x82, 0x82, 0x83, 0x83, 0x84,	/* 0x20 */
@@ -344,7 +348,7 @@ static u_char j2sjtable1[128] = {
 	0xe8, 0xe9, 0xe9, 0xea, 0xea, 0xeb, 0xeb, 0xec,	/* 0x70 */
 	0xec, 0xed, 0xed, 0xee, 0xee, 0xef, 0xef,    0
 };
-static u_char j2sjtable2[128] = {
+static CONST u_char j2sjtable2[128] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
 	   0, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46,	/* 0x20 */
@@ -360,7 +364,7 @@ static u_char j2sjtable2[128] = {
 	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x70 */
 	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e,    0
 };
-static u_char j2sjtable3[128] = {
+static CONST u_char j2sjtable3[128] = {
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x00 */
 	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	/* 0x10 */
 	   0, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5,	/* 0x20 */
@@ -459,46 +463,49 @@ static CONST kpathtable kpathlist[] = {
 	{&cappath, CAP},
 	{&utf8path, UTF8},
 	{&utf8macpath, M_UTF8},
+	{&utf8iconvpath, I_UTF8},
 	{&noconvpath, NOCNV},
 };
 #define	MAXKPATHLIST	arraysize(kpathlist)
 #endif	/* !_NOKANJIFCONV */
 
 #ifndef	FD
-int onkanji1 __P_((char *, int));
+int onkanji1 __P_((CONST char *, int));
 # if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
-int getlang __P_((char *, int));
+int getlang __P_((CONST char *, int));
 # endif
 # if	!defined (_NOENGMES) && !defined (_NOJPNMES)
-char *mesconv __P_((char *, char *));
+CONST char *mesconv __P_((CONST char *, CONST char *));
 # endif
 # if	!defined (_NOKANJICONV) \
 || (defined (FD) && defined (_USEDOSEMU) && defined (CODEEUC))
-int sjis2ujis __P_((char *, u_char *, int));
-int ujis2sjis __P_((char *, u_char *, int));
+int sjis2ujis __P_((char *, CONST u_char *, int));
+int ujis2sjis __P_((char *, CONST u_char *, int));
 # endif
 # ifdef	_USEUNICODE
 VOID readunitable __P_((int));
 VOID discardunitable __P_((VOID_A));
 u_int unifysjis __P_((u_int, int));
 u_int cnvunicode __P_((u_int, int));
+VOID ucs2normalization __P_((u_short *, int *, int, u_int, int));
+u_int ucs2denormalization __P_((CONST u_short *, int *, int));
 # endif
 # ifndef	_NOKANJICONV
 #  ifdef	_USEUNICODE
 int ucs2toutf8 __P_((char *, int, u_int));
-u_int ucs2fromutf8 __P_((u_char *, int));
+u_int ucs2fromutf8 __P_((CONST u_char *, int));
 #  endif
-int kanjiconv __P_((char *, char *, int, int, int, int));
-char *kanjiconv2 __P_((char *, char *, int, int, int, int));
-char *newkanjiconv __P_((char *, int, int, int));
+int kanjiconv __P_((char *, CONST char *, int, int, int, int));
+CONST char *kanjiconv2 __P_((char *, CONST char *, int, int, int, int));
+char *newkanjiconv __P_((CONST char *, int, int, int));
 int getoutputkcode __P_((VOID_A));
 # endif	/* !_NOKANJICONV */
 #endif	/* !FD */
 
 
 int onkanji1(s, ptr)
-char *s;
+CONST char *s;
 int ptr;
 {
 	int i;
@@ -520,8 +527,8 @@ int ptr;
 
 #if	!defined (_NOKANJICONV) \
 || (!defined (_NOENGMES) && !defined (_NOJPNMES))
-static char *NEAR strstr2(s1, s2)
-char *s1, *s2;
+static CONST char *NEAR strstr2(s1, s2)
+CONST char *s1, *s2;
 {
 	int i, c1, c2;
 
@@ -554,7 +561,7 @@ char *s1, *s2;
 
 /*ARGSUSED*/
 int getlang(s, io)
-char *s;
+CONST char *s;
 int io;
 {
 	int i, ret;
@@ -577,10 +584,13 @@ int io;
 #endif	/* !_NOKANJICONV || (!_NOENGMES && !_NOJPNMES) */
 
 #if	!defined (_NOENGMES) && !defined (_NOJPNMES)
-char *mesconv(jpn, eng)
-char *jpn, *eng;
+CONST char *mesconv(jpn, eng)
+CONST char *jpn, *eng;
 {
-	return((outputkcode == ENG) ? eng : jpn);
+	int n;
+
+	n = (messagelang != NOCNV) ? messagelang : outputkcode;
+	return((n == ENG) ? eng : jpn);
 }
 #endif
 
@@ -588,13 +598,13 @@ char *jpn, *eng;
 || (defined (FD) && defined (_USEDOSEMU) && defined (CODEEUC))
 static VOID NEAR sj2j(buf, s)
 char *buf;
-u_char *s;
+CONST u_char *s;
 {
 	u_int w;
 	int n, s1, s2, j1, j2, min, max;
 
-	s1 = s[0] & 0xff;
-	s2 = s[1] & 0xff;
+	s1 = (s[0] & 0xff);
+	s2 = (s[1] & 0xff);
 	if (s1 >= 0xf0) {
 		w = (((u_int)s1 << 8) | s2);
 		min = -1;
@@ -615,8 +625,8 @@ u_char *s;
 				break;
 			}
 		}
-		s1 = (w >> 8) & 0xff;
-		s2 = w & 0xff;
+		s1 = ((w >> 8) & 0xff);
+		s2 = (w & 0xff);
 	}
 	j1 = sj2jtable1[s1];
 	j2 = sj2jtable2[s2];
@@ -633,13 +643,13 @@ u_char *s;
 
 static VOID NEAR j2sj(buf, s)
 char *buf;
-u_char *s;
+CONST u_char *s;
 {
 	u_int w;
 	int i, s1, s2, j1, j2;
 
-	j1 = s[0] & 0x7f;
-	j2 = s[1] & 0x7f;
+	j1 = (s[0] & 0x7f);
+	j2 = (s[1] & 0x7f);
 
 	s1 = j2sjtable1[j1];
 	s2 = (j1 & 1) ? j2sjtable2[j2] : j2sjtable3[j2];
@@ -655,8 +665,8 @@ u_char *s;
 			continue;
 		w -= convtable[i].cnv;
 		w += convtable[i].start;
-		s1 = (w >> 8) & 0xff;
-		s2 = w & 0xff;
+		s1 = ((w >> 8) & 0xff);
+		s2 = (w & 0xff);
 		break;
 	}
 
@@ -666,7 +676,7 @@ u_char *s;
 
 int sjis2ujis(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 	int i, j;
@@ -691,7 +701,7 @@ int max;
 
 int ujis2sjis(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 	int i, j;
@@ -712,7 +722,7 @@ int max;
 
 #ifdef	_USEUNICODE
 static int NEAR openunitbl(file)
-char *file;
+CONST char *file;
 {
 	static int fd = -2;
 	u_char buf[2];
@@ -729,12 +739,20 @@ char *file;
 	if (!unitblpath || !*unitblpath) strcpy(path, file);
 	else strcatdelim2(path, unitblpath, file);
 
+# ifdef	FD
+	noconv++;
+# endif
+	fd = -1;
 	if ((fd = Xopen(path, O_BINARY | O_RDONLY, 0666)) < 0) fd = -1;
-	else if (!unitblent && sureread(fd, buf, 2) != 2) {
+	else if (unitblent) /*EMPTY*/;
+	else if (sureread(fd, buf, 2) == 2) unitblent = getword(buf, 0);
+	else {
 		Xclose(fd);
 		fd = -1;
 	}
-	else unitblent = getword(buf, 0);
+# ifdef	FD
+	noconv--;
+# endif
 
 	return(fd);
 }
@@ -911,7 +929,7 @@ int encode;
 				if ((wc & 0xff) < 0x80) return(wc);
 				break;
 			case 0xff00:
-				w = wc & 0xff;
+				w = (wc & 0xff);
 				if (w >= 0x21 && w <= 0x3a)
 					return(w + 0x8260 - 0x21);
 				if (w >= 0x41 && w <= 0x5a)
@@ -926,6 +944,8 @@ int encode;
 			kanjierrno = UTF8;
 			return(r);
 		}
+		if (wc == 0x3099) return(0x814a);
+		if (wc == 0x309a) return(0x814b);
 	}
 
 	if (unicodebuffer && !unitblbuf) readunitable(0);
@@ -999,7 +1019,7 @@ int encode;
 #ifndef	_NOKANJICONV
 # ifdef	_USEUNICODE
 static int NEAR opennftbl(file, nf, entp)
-char *file;
+CONST char *file;
 int nf;
 u_int *entp;
 {
@@ -1031,7 +1051,7 @@ u_int *entp;
 
 static int NEAR toenglish(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 	int i, j;
@@ -1052,7 +1072,7 @@ int max;
 
 static int NEAR tojis7(buf, s, max, knj, asc, io)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, knj, asc, io;
 {
 	int i, j, len, mode;
@@ -1073,8 +1093,8 @@ int max, knj, asc, io;
 			}
 			mode |= KANJI;
 # ifdef	CODEEUC
-			buf[j++] = s[i++] & ~0x80;
-			buf[j] = s[i] & ~0x80;
+			buf[j++] = (s[i++] & ~0x80);
+			buf[j] = (s[i] & ~0x80);
 # else
 			sj2j(&(buf[j++]), &(s[i++]));
 # endif
@@ -1087,7 +1107,7 @@ int max, knj, asc, io;
 			if (j + len > max) break;
 			if (!(mode & KANA)) buf[j++] = '\016';
 			mode |= KANA;
-			buf[j] = s[i] & ~0x80;
+			buf[j] = (s[i] & ~0x80);
 			buf[j] = jcnv(buf[j], io);
 		}
 		else {
@@ -1118,9 +1138,12 @@ int max, knj, asc, io;
 
 static int NEAR fromjis(buf, s, max, io)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, io;
 {
+# ifndef	CODEEUC
+	u_char tmp[MAXKLEN];
+# endif
 	int i, j, mode;
 
 	mode = ASCII;
@@ -1181,7 +1204,7 @@ int max, io;
 					if (j + 2 > max) break;
 					buf[j++] = (char)C_EKANA;
 # endif
-					buf[j++] = jdecnv(s[i], io) | 0x80;
+					buf[j++] = (jdecnv(s[i], io) | 0x80);
 				}
 			}
 			else if (mode & KANJI) {
@@ -1192,11 +1215,9 @@ int max, io;
 				}
 				else {
 # ifdef	CODEEUC
-					buf[j++] = s[i++] | 0x80;
-					buf[j++] = jdecnv(s[i], io) | 0x80;
+					buf[j++] = (s[i++] | 0x80);
+					buf[j++] = (jdecnv(s[i], io) | 0x80);
 # else
-					u_char tmp[2];
-
 					tmp[0] = s[i++];
 					tmp[1] = jdecnv(s[i], io);
 					j2sj(&(buf[j]), tmp);
@@ -1220,7 +1241,7 @@ int max, io;
 
 static int NEAR tojis8(buf, s, max, knj, asc, io)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, knj, asc, io;
 {
 	int i, j, len, mode;
@@ -1238,8 +1259,8 @@ int max, knj, asc, io;
 			}
 			mode = KANJI;
 # ifdef	CODEEUC
-			buf[j++] = s[i++] & ~0x80;
-			buf[j] = s[i] & ~0x80;
+			buf[j++] = (s[i++] & ~0x80);
+			buf[j] = (s[i] & ~0x80);
 # else
 			sj2j(&(buf[j++]), &(s[i++]));
 # endif
@@ -1272,7 +1293,7 @@ int max, knj, asc, io;
 
 static int NEAR tojunet(buf, s, max, knj, asc, io)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, knj, asc, io;
 {
 	int i, j, len, mode;
@@ -1290,8 +1311,8 @@ int max, knj, asc, io;
 			}
 			mode = KANJI;
 # ifdef	CODEEUC
-			buf[j++] = s[i++] & ~0x80;
-			buf[j] = s[i] & ~0x80;
+			buf[j++] = (s[i++] & ~0x80);
+			buf[j] = (s[i] & ~0x80);
 # else
 			sj2j(&(buf[j++]), &(s[i++]));
 # endif
@@ -1307,7 +1328,7 @@ int max, knj, asc, io;
 				buf[j++] = 'I';
 			}
 			mode = JKANA;
-			buf[j] = s[i] & ~0x80;
+			buf[j] = (s[i] & ~0x80);
 			buf[j] = jcnv(buf[j], io);
 		}
 		else {
@@ -1333,7 +1354,7 @@ int max, knj, asc, io;
 }
 
 # ifdef	_USEUNICODE
-static VOID NEAR ucs2normalization(buf, ptrp, max, wc, nf)
+VOID ucs2normalization(buf, ptrp, max, wc, nf)
 u_short *buf;
 int *ptrp, max;
 u_int wc;
@@ -1382,8 +1403,8 @@ int nf;
 	if (new) free(new);
 }
 
-static u_int NEAR ucs2denormalization(buf, ptrp, nf)
-u_short *buf;
+u_int ucs2denormalization(buf, ptrp, nf)
+CONST u_short *buf;
 int *ptrp, nf;
 {
 	u_char *cp, *new;
@@ -1454,15 +1475,16 @@ int *ptrp, nf;
 }
 
 static u_int NEAR toucs2(s, ptrp)
-u_char *s;
+CONST u_char *s;
 int *ptrp;
 {
+# ifdef	CODEEUC
+	u_char tmp[MAXKLEN];
+# endif
 	u_int w;
 
 	if (iskanji1((char *)s, *ptrp)) {
 # ifdef	CODEEUC
-		u_char tmp[2];
-
 		j2sj((char *)tmp, &(s[*ptrp]));
 		if (kanjierrno) kanjierrno = EUC;
 		*ptrp += 2;
@@ -1485,11 +1507,14 @@ char *buf;
 int *ptrp;
 u_int wc;
 {
+# ifdef	CODEEUC
+	u_char tmp[MAXKLEN];
+# endif
 	int c1, c2;
 
 	wc = cnvunicode(wc, 0);
-	c1 = (wc >> 8) & 0xff;
-	c2 = wc & 0xff;
+	c1 = ((wc >> 8) & 0xff);
+	c2 = (wc & 0xff);
 	if (wc > 0xa0 && wc <= 0xdf) {
 # ifdef	CODEEUC
 		buf[(*ptrp)++] = (char)C_EKANA;
@@ -1498,8 +1523,6 @@ u_int wc;
 	}
 	else if (issjis1(c1) && issjis2(c2)) {
 # ifdef	CODEEUC
-		u_char tmp[2];
-
 		tmp[0] = c1;
 		tmp[1] = c2;
 		sj2j(&(buf[*ptrp]), tmp);
@@ -1520,20 +1543,20 @@ u_int wc;
 {
 	if (wc < 0x80) buf[ptr++] = wc;
 	else if (wc < 0x800) {
-		buf[ptr++] = 0xc0 | (wc >> 6);
-		buf[ptr++] = 0x80 | (wc & 0x3f);
+		buf[ptr++] = (0xc0 | (wc >> 6));
+		buf[ptr++] = (0x80 | (wc & 0x3f));
 	}
 	else {
-		buf[ptr++] = 0xe0 | (wc >> 12);
-		buf[ptr++] = 0x80 | ((wc >> 6) & 0x3f);
-		buf[ptr++] = 0x80 | (wc & 0x3f);
+		buf[ptr++] = (0xe0 | (wc >> 12));
+		buf[ptr++] = (0x80 | ((wc >> 6) & 0x3f));
+		buf[ptr++] = (0x80 | (wc & 0x3f));
 	}
 
 	return(ptr);
 }
 
 u_int ucs2fromutf8(s, ptrp)
-u_char *s;
+CONST u_char *s;
 int *ptrp;
 {
 	u_int w;
@@ -1542,10 +1565,11 @@ int *ptrp;
 	ptr = (ptrp) ? *ptrp : 0;
 	w = s[ptr++];
 	if (w < 0x80) /*EMPTY*/;
-	else if (isutf2(w, s[ptr])) w = ((w & 0x1f) << 6) | (s[ptr++] & 0x3f);
+	else if (isutf2(w, s[ptr]))
+		w = (((w & 0x1f) << 6) | (s[ptr++] & 0x3f));
 	else if (isutf3(w, s[ptr], s[ptr + 1])) {
-		w = ((w & 0x0f) << 6) | (s[ptr++] & 0x3f);
-		w = (w << 6) | (s[ptr++] & 0x3f);
+		w = (((w & 0x0f) << 6) | (s[ptr++] & 0x3f));
+		w = ((w << 6) | (s[ptr++] & 0x3f));
 	}
 	else {
 		w = U2_UDEF;
@@ -1558,7 +1582,7 @@ int *ptrp;
 
 static int NEAR toutf8(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 	u_int c;
@@ -1581,7 +1605,7 @@ int max;
 
 static int NEAR fromutf8(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 	int i, j;
@@ -1595,7 +1619,7 @@ int max;
 
 static int NEAR toutf8nf(buf, s, max, nf)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, nf;
 {
 	u_short *u1, *u2;
@@ -1605,10 +1629,10 @@ int max, nf;
 	u2 = (u_short *)malloc2((max + 1) * sizeof(u_short));
 
 	for (i = j = 0; s[i] && j < max; j++) u1[j] = toucs2(s, &i);
-	u1[j] = 0;
+	u1[j] = (u_short)0;
 	for (i = j = 0; u1[i] && j < max; i++)
 		ucs2normalization(u2, &j, max, u1[i], nf);
-	u2[j] = 0;
+	u2[j] = (u_short)0;
 	for (i = j = 0; u2[i] && j < max; i++) {
 		len = 1;
 		if (u2[i] >= 0x80) len++;
@@ -1627,7 +1651,7 @@ int max, nf;
 
 static int NEAR fromutf8nf(buf, s, max, nf)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max, nf;
 {
 	u_short *u1, *u2;
@@ -1637,10 +1661,10 @@ int max, nf;
 	u2 = (u_short *)malloc2((max + 1) * sizeof(u_short));
 
 	for (i = j = 0; s[i] && j < max; j++) u1[j] = ucs2fromutf8(s, &i);
-	u1[j] = 0;
+	u1[j] = (u_short)0;
 	for (i = j = 0; u1[i] && j < max; j++)
 		u2[j] = ucs2denormalization(u1, &i, nf);
-	u2[j] = 0;
+	u2[j] = (u_short)0;
 	for (i = j = 0; u2[i] && j + 1 < max; i++) fromucs2(buf, &j, u2[i]);
 
 	free(u1);
@@ -1667,11 +1691,11 @@ int c;
 
 static int NEAR tohex(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 # ifdef	CODEEUC
-	u_char tmp[2];
+	u_char tmp[MAXKLEN];
 # endif
 	int i, j;
 
@@ -1700,9 +1724,12 @@ int max;
 
 static int NEAR fromhex(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
+# ifdef	CODEEUC
+	u_char tmp[MAXKLEN];
+# endif
 	int i, j, c1, c2;
 
 	for (i = j = 0; s[i] && j < max; i++, j++) {
@@ -1731,8 +1758,6 @@ int max;
 				}
 				else {
 # ifdef	CODEEUC
-					u_char tmp[2];
-
 					tmp[0] = c1;
 					tmp[1] = c2;
 					sj2j(&(buf[j]), tmp);
@@ -1770,11 +1795,11 @@ int c;
 
 static int NEAR tocap(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
 # ifdef	CODEEUC
-	u_char tmp[2];
+	u_char tmp[MAXKLEN];
 # endif
 	int i, j, len;
 
@@ -1807,9 +1832,12 @@ int max;
 
 static int NEAR fromcap(buf, s, max)
 char *buf;
-u_char *s;
+CONST u_char *s;
 int max;
 {
+# ifdef	CODEEUC
+	u_char tmp[MAXKLEN];
+# endif
 	int i, j, c1, c2;
 
 	for (i = j = 0; s[i] && j < max; i++, j++) {
@@ -1838,8 +1866,6 @@ int max;
 				}
 				else {
 # ifdef	CODEEUC
-					u_char tmp[2];
-
 					tmp[0] = c1;
 					tmp[1] = c2;
 					sj2j(&(buf[j]), tmp);
@@ -1858,8 +1884,9 @@ int max;
 	return(j);
 }
 
-static char *NEAR _kanjiconv(buf, s, max, in, out, lenp, io)
-char *buf, *s;
+static CONST char *NEAR _kanjiconv(buf, s, max, in, out, lenp, io)
+char *buf;
+CONST char *s;
 int max, in, out, *lenp, io;
 {
 	kanjierrno = 0;
@@ -1892,6 +1919,7 @@ int max, in, out, *lenp, io;
 			*lenp = toutf8(buf, (u_char *)s, max);
 			break;
 		case M_UTF8:
+		case I_UTF8:
 			*lenp = toutf8nf(buf, (u_char *)s, max, out - UTF8);
 			break;
 # endif
@@ -1921,6 +1949,7 @@ int max, in, out, *lenp, io;
 						max);
 					break;
 				case M_UTF8:
+				case I_UTF8:
 					*lenp = fromutf8nf(buf, (u_char *)s,
 						max, in - UTF8);
 					break;
@@ -1951,7 +1980,8 @@ int max, in, out, *lenp, io;
 }
 
 int kanjiconv(buf, s, max, in, out, io)
-char *buf, *s;
+char *buf;
+CONST char *s;
 int max, in, out, io;
 {
 	int len;
@@ -1963,8 +1993,9 @@ int max, in, out, io;
 	return(len);
 }
 
-char *kanjiconv2(buf, s, max, in, out, io)
-char *buf, *s;
+CONST char *kanjiconv2(buf, s, max, in, out, io)
+char *buf;
+CONST char *s;
 int max, in, out, io;
 {
 	int len;
@@ -1976,18 +2007,18 @@ int max, in, out, io;
 }
 
 char *newkanjiconv(s, in, out, io)
-char *s;
+CONST char *s;
 int in, out, io;
 {
 	char *buf;
 	int len;
 
-	if (!s) return(s);
+	if (!s) return(NULL);
 	len = strlen(s) * 3 + 3;
-	if (!(buf = malloc(len + 1))) return(s);
+	if (!(buf = malloc(len + 1))) return((char *)s);
 	if (kanjiconv2(buf, s, len, in, out, io) != buf) {
 		free(buf);
-		return(s);
+		return((char *)s);
 	}
 
 	return(buf);
@@ -1996,7 +2027,7 @@ int in, out, io;
 
 #if	defined (FD) && !defined (_NOKANJIFCONV)
 int getkcode(path)
-char *path;
+CONST char *path;
 {
 	int i;
 
@@ -2034,7 +2065,12 @@ int dos;
 # ifndef	_NOKANJIFCONV
 	int fgetok;
 # endif
-	char *cp;
+	CONST char *cp;
+
+#ifdef	DOUBLESLASH
+	if (path[0] == _SC_ && path[1] == _SC_ && !isdslash(path))
+		memmove(path, &(path[1]), strlen(&(path[1])) + 1);
+#endif
 
 	if (noconv) return(path);
 # ifndef	_NOKANJIFCONV
@@ -2068,12 +2104,13 @@ int dos;
 	}
 # endif
 
-	return(cp);
+	return((char *)cp);
 }
 
 /*ARGSUSED*/
 char *convput(buf, path, needfile, rdlink, rrreal, codep)
-char *buf, *path;
+char *buf;
+CONST char *path;
 int needfile, rdlink;
 char *rrreal;
 int *codep;
@@ -2087,7 +2124,8 @@ int *codep;
 # ifndef	_NOKANJIFCONV
 	int c, fputok;
 # endif
-	char *cp, *file, rpath[MAXPATHLEN];
+	CONST char *cp, *file;
+	char *tmp, rpath[MAXPATHLEN];
 	int n;
 
 	if (rrreal) *rrreal = '\0';
@@ -2097,7 +2135,7 @@ int *codep;
 # ifdef	_USEDOSEMU
 		if (dospath(path, buf)) return(buf);
 # endif
-		return(path);
+		return((char *)path);
 	}
 # ifndef	_NOKANJIFCONV
 	fputok = (nokanjifconv) ? 0 : 1;
@@ -2107,13 +2145,20 @@ int *codep;
 	if (norealpath) cp = path;
 	else {
 		if ((file = strrdelim(path, 0))) {
-			n = file - path;
-			if (file++ == isrootdir(path)) n++;
+# ifdef	DOUBLESLASH
+			if ((n = isdslash(path)) && file < &(path[n]))
+				file = &(path[n]);
+			else
+# endif
+			{
+				n = file - path;
+				if (file++ == isrootdir(path)) n++;
+			}
 			strncpy2(rpath, path, n);
 		}
 # ifdef	_USEDOSEMU
 		else if ((n = _dospath(path))) {
-			file = path + 2;
+			file = &(path[2]);
 			VOID_C gendospath(rpath, n, '.');
 		}
 # endif
@@ -2122,10 +2167,11 @@ int *codep;
 			copycurpath(rpath);
 		}
 		realpath2(rpath, rpath, 1);
-		cp = strcatdelim(rpath);
-		strncpy2(cp, file, MAXPATHLEN - 1 - (cp - rpath));
+		tmp = strcatdelim(rpath);
+		strncpy2(tmp, file, MAXPATHLEN - 1 - (tmp - rpath));
 		cp = rpath;
 	}
+
 # ifdef	_USEDOSEMU
 	if ((n = dospath(cp, kbuf))) {
 		cp = kbuf;
@@ -2147,11 +2193,11 @@ int *codep;
 		if (rrreal) strcpy(rrreal, rbuf);
 		if (rdlink && rrreadlink(cp, buf, MAXPATHLEN - 1) >= 0) {
 			if (needfile && strdelim(buf, 0)) needfile = 0;
-			if (*buf == _SC_ || !(cp = strrdelim(rbuf, 0)))
+			if (*buf == _SC_ || !(tmp = strrdelim(rbuf, 0)))
 				cp = buf;
 			else {
-				cp++;
-				strncpy2(cp, buf, MAXPATHLEN - (cp - rbuf));
+				tmp++;
+				strncpy2(tmp, buf, MAXPATHLEN - (tmp - rbuf));
 				cp = rbuf;
 			}
 			realpath2(cp, rpath, 1);
@@ -2159,14 +2205,16 @@ int *codep;
 		}
 	}
 # endif
-	if (cp == path) return(path);
+	if (cp == path) return((char *)path);
 	if (needfile && (file = strrdelim(cp, 0))) file++;
 	else file = cp;
+
 # ifdef	_USEDOSEMU
-	if (n && !_dospath(file)) strcpy(gendospath(buf, n, '\0'), file);
+	if (isalpha2(n) && !_dospath(file)) tmp = gendospath(buf, n, '\0');
 	else
 # endif
-	strcpy(buf, file);
+	tmp = buf;
+	strcpy(tmp, file);
 
 	return(buf);
 }
