@@ -16,7 +16,7 @@
 #if	MSDOS
 #include <sys/timeb.h>
 extern int setcurdrv __P_((int, int));
-extern char *unixrealpath __P_((char *, char *));
+extern char *unixrealpath __P_((CONST char *, char *));
 #endif
 
 #ifdef	_NOORIGSHELL
@@ -43,26 +43,23 @@ extern char **environ;
 extern char fullpath[];
 extern char *origpath;
 extern int hideclock;
-#ifndef	_NODOSDRIVE
-extern int lastdrv;
-#endif
 
 #ifndef	NOSYMLINK
 static int NEAR evallink __P_((char *, char *));
 #endif
-static char *NEAR _realpath2 __P_((char *, char *, int));
+static char *NEAR _realpath2 __P_((CONST char *, char *, int));
 #ifdef	_NOORIGSHELL
-static int NEAR _getenv2 __P_((char *, int, char **));
+static int NEAR _getenv2 __P_((CONST char *, int, char **));
 static char **NEAR _putenv2 __P_((char *, char **));
 #endif
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
-static long NEAR char2long __P_((u_char *));
-static int NEAR tmcmp __P_((struct tm *, struct tm *));
+static long NEAR char2long __P_((CONST u_char *));
+static int NEAR tmcmp __P_((CONST struct tm *, CONST struct tm *));
 #endif
 #if	!defined (USEMKTIME) && !defined (USETIMELOCAL)
-static time_t NEAR timegm2 __P_((struct tm *));
-static long NEAR gettimezone __P_((struct tm *, time_t));
+static time_t NEAR timegm2 __P_((CONST struct tm *));
+static long NEAR gettimezone __P_((CONST struct tm *, time_t));
 #endif
 
 #ifdef	_NOORIGSHELL
@@ -79,7 +76,7 @@ static int wasttyflags = 0;
 
 
 int stat2(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 #ifdef	NOSYMLINK
@@ -142,7 +139,8 @@ char *path, *delim;
 
 /*ARGSUSED*/
 static char *NEAR _realpath2(path, resolved, rdlink)
-char *path, *resolved;
+CONST char *path;
+char *resolved;
 int rdlink;
 {
 	char *cp, *top;
@@ -160,16 +158,23 @@ int rdlink;
 	if (n == 1) {
 		cp = strrdelim(resolved, 0);
 		top = resolved;
-#if	MSDOS
-		top += 2;
-#else	/* !MSDOS */
-# ifndef	_NODOSDRIVE
+#ifdef	_USEDOSEMU
 		if (_dospath(resolved)) top += 2;
 		else
-# endif
+#endif
+#ifdef	DOUBLESLASH
+		if ((n = isdslash(resolved)) && cp < &(resolved[n])) {
+			top = &(resolved[n - 1]);
+			if (*top != _SC_) return(resolved);
+		}
+		else
+#endif
+#if	MSDOS
+		top += 2;
+#else
 		if (rdlink && evallink(resolved, cp) > 0)
 			return(_realpath2(path, resolved, rdlink));
-#endif	/* !MSDOS */
+#endif
 		if (!cp || cp == top) {
 			*top = _SC_;
 			cp = top + 1;
@@ -192,16 +197,20 @@ int rdlink;
 }
 
 char *realpath2(path, resolved, rdlink)
-char *path, *resolved;
+CONST char *path;
+char *resolved;
 int rdlink;
 {
-#if	!MSDOS && !defined (_NODOSDRIVE)
+#ifdef	_USEDOSEMU
 	char *cp;
 	int duplastdrive;
 #endif
-#if	MSDOS || !defined (_NODOSDRIVE)
+#ifdef	_USEDOSPATH
 	int drv;
 #endif
+#ifdef	DOUBLESLASH
+	int n;
+#endif
 #if	MSDOS
 	int drive;
 #endif
@@ -210,25 +219,8 @@ int rdlink;
 	strcpy(tmp, path);
 	path = tmp;
 
-#if	MSDOS
-	drv = dospath(nullstr, NULL);
-	if ((drive = _dospath(path))) path += 2;
-	if (*path == _SC_) {
-		if (!drive) drive = drv;
-		VOID_C gendospath(resolved, drive, _SC_);
-	}
-	else if (drive && drive != drv) {
-		if (setcurdrv(drive, 0) < 0)
-			VOID_C gendospath(resolved, drive, _SC_);
-		else {
-			if (!Xgetwd(resolved)) lostcwd(resolved);
-			if (setcurdrv(drv, 0) < 0) error("setcurdrv()");
-		}
-	}
-#else	/* !MSDOS */
-	if (*path == _SC_) copyrootpath(resolved);
-# ifndef	_NODOSDRIVE
-	else if ((drv = _dospath(path))) {
+#ifdef	_USEDOSEMU
+	if ((drv = _dospath(path))) {
 		path += 2;
 		if (*path == _SC_) cp = NULL;
 		else {
@@ -239,11 +231,40 @@ int rdlink;
 		}
 		if (!cp) VOID_C gendospath(resolved, drv, _SC_);
 	}
-# endif
+	else
+#endif
+#ifdef	DOUBLESLASH
+	if ((n = isdslash(path))) {
+		strncpy2(resolved, path, n);
+		path += n;
+		if (*path == _SC_) path++;
+	}
+	else
+#endif
+	{
+#if	MSDOS
+		drv = dospath(nullstr, NULL);
+		if ((drive = _dospath(path))) path += 2;
+		if (*path == _SC_) {
+			if (!drive) drive = drv;
+			VOID_C gendospath(resolved, drive, _SC_);
+		}
+		else if (drive && drive != drv) {
+			if (setcurdrv(drive, 0) < 0)
+				VOID_C gendospath(resolved, drive, _SC_);
+			else {
+				if (!Xgetwd(resolved)) lostcwd(resolved);
+				if (setcurdrv(drv, 0) < 0) error("setcurdrv()");
+			}
+		}
+#else	/* !MSDOS */
+		if (*path == _SC_) copyrootpath(resolved);
 #endif	/* !MSDOS */
-	else if (!rdlink && resolved != fullpath && *fullpath)
-		strcpy(resolved, fullpath);
-	else if (!Xgetwd(resolved)) copyrootpath(resolved);
+		else if (!rdlink && resolved != fullpath && *fullpath)
+			strcpy(resolved, fullpath);
+		else if (!Xgetwd(resolved)) copyrootpath(resolved);
+	}
+
 	norealpath++;
 	_realpath2(path, resolved, rdlink);
 	norealpath--;
@@ -252,7 +273,7 @@ int rdlink;
 }
 
 int _chdir2(path)
-char *path;
+CONST char *path;
 {
 	char cwd[MAXPATHLEN];
 
@@ -296,7 +317,7 @@ char *path;
 }
 
 int chdir2(path)
-char *path;
+CONST char *path;
 {
 	char cwd[MAXPATHLEN], tmp[MAXPATHLEN];
 	int duperrno;
@@ -340,7 +361,7 @@ char *path;
 		unixpath = NULL;
 	}
 #endif
-	if (getconstvar("PWD")) setenv2("PWD", fullpath, 1);
+	if (getconstvar(ENVPWD)) setenv2(ENVPWD, fullpath, 1);
 #if	MSDOS
 	if (unixrealpath(fullpath, tmp)) strcpy(fullpath, tmp);
 #endif
@@ -353,13 +374,13 @@ char *path;
 }
 
 int chdir3(path, raw)
-char *path;
+CONST char *path;
 int raw;
 {
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
-	char *cwd;
+	CONST char *cwd;
 
 	cwd = path;
 	if (!raw && path[0] && !path[1]) switch (path[0]) {
@@ -412,7 +433,7 @@ int mode;
 {
 	char *cp1, *cp2, *eol;
 
-	eol = path + (int)strlen(path) - 1;
+	eol = &(path[(int)strlen(path) - 1]);
 	while (eol > path && *eol == _SC_) eol--;
 #ifdef	BSPATHDELIM
 	if (onkanji1(path, eol - path)) eol++;
@@ -488,7 +509,7 @@ ALLOC_T n, *sizep;
 }
 
 char *strdup2(s)
-char *s;
+CONST char *s;
 {
 	char *tmp;
 	int n;
@@ -502,7 +523,7 @@ char *s;
 }
 
 char *strndup2(s, n)
-char *s;
+CONST char *s;
 int n;
 {
 	char *tmp;
@@ -518,13 +539,13 @@ int n;
 }
 
 char *strchr2(s, c)
-char *s;
+CONST char *s;
 int c;
 {
 	int i;
 
 	for (i = 0; s[i]; i++) {
-		if (s[i] == c) return(&(s[i]));
+		if (s[i] == c) return((char *)&(s[i]));
 		if (iskanji1(s, i)) i++;
 #ifdef	CODEEUC
 		else if (isekana(s, i)) i++;
@@ -535,7 +556,7 @@ int c;
 }
 
 char *strrchr2(s, c)
-char *s;
+CONST char *s;
 int c;
 {
 	int i;
@@ -543,7 +564,7 @@ int c;
 
 	cp = NULL;
 	for (i = 0; s[i]; i++) {
-		if (s[i] == c) cp = &(s[i]);
+		if (s[i] == c) cp = (char *)&(s[i]);
 		if (iskanji1(s, i)) i++;
 #ifdef	CODEEUC
 		else if (isekana(s, i)) i++;
@@ -554,7 +575,8 @@ int c;
 }
 
 char *strcpy2(s1, s2)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 {
 	int i;
 
@@ -565,7 +587,8 @@ char *s1, *s2;
 }
 
 char *strncpy2(s1, s2, n)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 int n;
 {
 	int i;
@@ -581,7 +604,8 @@ int n;
  *	strncpy3(buf, s, &(-x), 0): same as sprintf(buf, "%s", s);
  */
 int strncpy3(s1, s2, lenp, ptr)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 int *lenp, ptr;
 {
 	int i, j, len;
@@ -638,7 +662,7 @@ int *lenp, ptr;
 
 #ifdef	CODEEUC
 int strlen2(s)
-char *s;
+CONST char *s;
 {
 	int i, len;
 
@@ -649,7 +673,7 @@ char *s;
 #endif	/* CODEEUC */
 
 int strlen3(s)
-char *s;
+CONST char *s;
 {
 	int i, len;
 
@@ -671,7 +695,7 @@ char *s;
 }
 
 int atoi2(s)
-char *s;
+CONST char *s;
 {
 	int n;
 
@@ -703,7 +727,7 @@ va_dcl
 }
 
 VOID perror2(s)
-char *s;
+CONST char *s;
 {
 	int duperrno;
 
@@ -716,7 +740,7 @@ char *s;
 
 #ifdef	_NOORIGSHELL
 static int NEAR _getenv2(name, len, envp)
-char *name;
+CONST char *name;
 int len;
 char **envp;
 {
@@ -758,12 +782,16 @@ char *s, **envp;
 #endif	/* _NOORIGSHELL */
 
 char *getenv2(name)
-char *name;
+CONST char *name;
 {
 #ifdef	_NOORIGSHELL
 	char **envpp[2];
 	int i, n, len;
+#else
+	char *cp;
+#endif
 
+#ifdef	_NOORIGSHELL
 	len = strlen(name);
 	envpp[0] = environ2;
 	envpp[1] = environ;
@@ -777,8 +805,6 @@ char *name;
 			return(&(envpp[i][n][len - FDESIZ + 1]));
 	}
 #else	/* !_NOORIGSHELL */
-	char *cp;
-
 	if ((cp = getshellvar(name, -1))) return(cp);
 	if (!strnenvcmp(name, FDENV, FDESIZ)
 	&& (cp = getshellvar(&(name[FDESIZ]), -1)))
@@ -789,7 +815,7 @@ char *name;
 }
 
 int setenv2(name, value, export)
-char *name, *value;
+CONST char *name, *value;
 int export;
 {
 	char *cp;
@@ -801,7 +827,7 @@ int export;
 	len = strlen(name);
 	if (!value) {
 #ifdef	_NOORIGSHELL
-		cp = name;
+		cp = (char *)name;
 #else
 		return(unset(name, len));
 #endif
@@ -850,7 +876,7 @@ sigcst_t func;
 #endif	/* USESIGACTION */
 
 int system2(command, flags)
-char *command;
+CONST char *command;
 int flags;
 {
 	int n, wastty, mode, ret;
@@ -881,7 +907,7 @@ int flags;
 		fflush(stderr);
 		Xttyiomode(1);
 		keyflush();
-		getkey2(0);
+		getkey3(0, inputkcode);
 		Xstdiomode();
 		fputnl(stderr);
 	}
@@ -900,7 +926,7 @@ int flags;
 }
 
 FILE *popen2(command)
-char *command;
+CONST char *command;
 {
 	FILE *fp;
 	int n;
@@ -929,7 +955,7 @@ char *command;
 		fflush(stderr);
 		Xttyiomode(1);
 		keyflush();
-		getkey2(0);
+		getkey3(0, inputkcode);
 		Xstdiomode();
 		fputnl(stderr);
 		if (wasttyflags & F_TTYIOMODE)
@@ -960,10 +986,28 @@ char *getwd2(VOID_A)
 	return(strdup2(cwd));
 }
 
+time_t time2(VOID_A)
+{
+#if	MSDOS
+	struct timeb buffer;
+
+	ftime(&buffer);
+
+	return((time_t)(buffer.time));
+#else
+	struct timeval t_val;
+	struct timezone tz;
+
+	gettimeofday2(&t_val, &tz);
+
+	return((time_t)(t_val.tv_sec));
+#endif
+}
+
 #if	!MSDOS && !defined (NOTZFILEH) \
 && !defined (USEMKTIME) && !defined (USETIMELOCAL)
 static long NEAR char2long(s)
-u_char *s;
+CONST u_char *s;
 {
 	return((long)((u_long)(s[3])
 		| ((u_long)(s[2]) << (BITSPERBYTE * 1))
@@ -972,7 +1016,7 @@ u_char *s;
 }
 
 static int NEAR tmcmp(tm1, tm2)
-struct tm *tm1, *tm2;
+CONST struct tm *tm1, *tm2;
 {
 	if (tm1 -> tm_year != tm2 -> tm_year)
 		return (tm1 -> tm_year - tm2 -> tm_year);
@@ -991,7 +1035,7 @@ struct tm *tm1, *tm2;
 
 #if	!defined (USEMKTIME) && !defined (USETIMELOCAL)
 static time_t NEAR timegm2(tm)
-struct tm *tm;
+CONST struct tm *tm;
 {
 	time_t t;
 	int i, y;
@@ -1025,7 +1069,7 @@ struct tm *tm;
 }
 
 static long NEAR gettimezone(tm, t)
-struct tm *tm;
+CONST struct tm *tm;
 time_t t;
 {
 # if	MSDOS
@@ -1132,24 +1176,6 @@ time_t t;
 }
 #endif	/* !USEMKTIME && !USETIMELOCAL */
 
-time_t time2(VOID_A)
-{
-#if	MSDOS
-	struct timeb buffer;
-
-	ftime(&buffer);
-
-	return((time_t)(buffer.time));
-#else
-	struct timeval t_val;
-	struct timezone tz;
-
-	gettimeofday2(&t_val, &tz);
-
-	return((time_t)(t_val.tv_sec));
-#endif
-}
-
 time_t timelocal2(tm)
 struct tm *tm;
 {
