@@ -26,12 +26,14 @@
 # define	SCRWIDTH	0x4a
 # endif
 # ifdef	DJGPP
+# define	__term_regs	__dpmi_regs
 # define	intdos2(rp)	__dpmi_int(0x21, rp)
 # define	getkeybuf(o)	_farpeekw(_dos_ds, KEYBUFWORKSEG * 0x10 + o)
 # define	putkeybuf(o, n)	_farpokew(_dos_ds, KEYBUFWORKSEG * 0x10 + o, n)
 # define	intbios(v, rp)	__dpmi_int(v, rp)
 # define	getbiosbyte(o)	_farpeekb(_dos_ds, BIOSSEG * 0x10 + o)
 # else	/* !DJGPP */
+typedef union REGS	__term_regs;
 # define	intdos2(rp)	int86(0x21, rp, rp)
 # define	getkeybuf(o)	(*((u_short far *)MK_FP(KEYBUFWORKSEG, o)))
 # define	putkeybuf(o, n)	(*((u_short far *)MK_FP(KEYBUFWORKSEG, o)) = n)
@@ -78,11 +80,17 @@ extern int errno;
 #define	strerror2		strerror
 #else
 # ifndef	DECLERRLIST
-extern char *sys_errlist[];
+extern CONST char *CONST sys_errlist[];
 # endif
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
 
+#ifdef	_POSIX_VDISABLE
+#define	K_UNDEF		_POSIX_VDISABLE
+#else
+#define	K_UNDEF		0xff
+#endif
+
 #ifdef	DEBUG
 extern VOID muntrace __P_ ((VOID_A));
 extern char *_mtrace_file;
@@ -207,11 +215,11 @@ typedef int	tputs_t;
 # define	TERM_at1		key_beg
 # define	TERM_at7		key_end
 # else	/* !USETERMINFO */
-extern int tgetent __P_((char *, char *));
-extern int tgetnum __P_((char *));
-extern int tgetflag __P_((char *));
-extern char *tgetstr __P_((char *, char **));
-extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
+extern int tgetent __P_((char *, CONST char *));
+extern int tgetnum __P_((CONST char *));
+extern int tgetflag __P_((CONST char *));
+extern char *tgetstr __P_((CONST char *, char **));
+extern int tputs __P_((CONST char *, int, int (*)__P_((tputs_t))));
 # define	tgetnum2		tgetnum
 # define	tgetflag2		tgetflag
 # define	TERM_pc			"pc"
@@ -334,9 +342,9 @@ extern int tputs __P_((char *, int, int (*)__P_((tputs_t))));
 #endif
 #endif	/* !MSDOS */
 
-static int NEAR err2 __P_((char *));
+static int NEAR err2 __P_((CONST char *));
 #if	!MSDOS
-static char *NEAR tstrdup __P_((char *));
+static char *NEAR tstrdup __P_((CONST char *));
 #endif
 static int NEAR defaultterm __P_((VOID_A));
 static int NEAR maxlocate __P_((int *, int *));
@@ -347,7 +355,7 @@ static int NEAR biosscroll __P_((int, int, int, int, int));
 static int NEAR biosputch __P_((int, int));
 static int NEAR bioscurstype __P_((int));
 static int NEAR chgattr __P_((int));
-static int NEAR evalcsi __P_((char *));
+static int NEAR evalcsi __P_((CONST char *));
 # endif
 # if	!defined (DJGPP) || defined (NOTUSEBIOS) || defined (PC98)
 static int NEAR dosgettime __P_((u_char []));
@@ -358,14 +366,15 @@ static int NEAR ttymode __P_((int, int, int, int));
 # else
 static int NEAR ttymode __P_((int, int, int, int, int, int, int, int));
 # endif
-static char *NEAR tgetstr2 __P_((char **, char *));
-static char *NEAR tgetstr3 __P_((char **, char *, char *));
-static char *NEAR tgetkeyseq __P_((int, char *));
+static char *NEAR tgetstr2 __P_((char **, CONST char *));
+static char *NEAR tgetstr3 __P_((char **, CONST char *, CONST char *));
+static char *NEAR tgetkeyseq __P_((int, CONST char *));
 static kstree_t *NEAR newkeyseqtree __P_((kstree_t *, int));
 static int NEAR freekeyseqtree __P_((kstree_t *, int));
 static int cmpkeyseq __P_((CONST VOID_P, CONST VOID_P));
 static int NEAR sortkeyseq __P_((VOID_A));
 static int putch3 __P_((tputs_t));
+static int NEAR ungetch2 __P_((int));
 #endif	/* !MSDOS */
 
 #if	!MSDOS && !defined (USETERMINFO)
@@ -383,7 +392,11 @@ int n_column = -1;
 int n_lastcolumn = -1;
 int n_line = -1;
 int stable_standout = 0;
+#if	MSDOS
+CONST char *termstr[MAXTERMSTR];
+#else
 char *termstr[MAXTERMSTR];
+#endif
 u_char cc_intr = K_CTRL('C');
 u_char cc_quit = K_CTRL('\\');
 #if	MSDOS
@@ -391,7 +404,7 @@ u_char cc_eof = K_CTRL('Z');
 #else
 u_char cc_eof = K_CTRL('D');
 #endif
-u_char cc_eol = 255;
+u_char cc_eol = K_UNDEF;
 u_char cc_erase = K_CTRL('H');
 int (*keywaitfunc)__P_((VOID_A)) = NULL;
 #if	!MSDOS
@@ -437,7 +450,7 @@ static u_short keybuftop = 0;
 # ifdef	USEVIDEOBIOS
 static u_short videoattr = 0x07;
 # endif
-static int specialkeycode[] = {
+static CONST int specialkeycode[] = {
 	0,
 	K_UP, K_DOWN, K_RIGHT, K_LEFT,
 	K_IC, K_DC, K_HOME, K_END, K_PPAGE, K_NPAGE,
@@ -448,11 +461,11 @@ static int specialkeycode[] = {
 };
 #define	SPECIALKEYSIZ	arraysize(specialkeycode)
 #else	/* !MSDOS */
-static char *dumblist[] = {"dumb", "un", "unknown"};
+static CONST char *dumblist[] = {"dumb", "un", "unknown"};
 #define	DUMBLISTSIZE	arraysize(dumblist)
 static keyseq_t keyseq[K_MAX - K_MIN + 1];
 static kstree_t *keyseqtree = NULL;
-static char *defkeyseq[K_MAX - K_MIN + 1] = {
+static CONST char *defkeyseq[K_MAX - K_MIN + 1] = {
 	NULL,			/* K_NOKEY */
 	"\033OB",		/* K_DOWN */
 	"\033OA",		/* K_UP */
@@ -569,12 +582,8 @@ static char *defkeyseq[K_MAX - K_MIN + 1] = {
 	NULL,			/* K_HELP */
 };
 #endif	/* !MSDOS */
-
-#if	MSDOS || !defined (TIOCSTI)
-static u_char ungetbuf[10];
+static u_char ungetbuf[16];
 static int ungetnum = 0;
-#endif
-
 static int termflags = 0;
 #define	F_INITTTY	001
 #define	F_TERMENT	002
@@ -582,7 +591,7 @@ static int termflags = 0;
 #define	F_TTYCHANGED	010
 #define	F_INPROGRESS	020
 #define	F_RESETTTY	(F_INITTTY | F_TTYCHANGED)
-static char *deftermstr[MAXTERMSTR] = {
+static CONST char *deftermstr[MAXTERMSTR] = {
 #ifdef	PC98
 	"\033[>1h",		/* T_INIT */
 	"\033[>1l",		/* T_END */
@@ -815,7 +824,7 @@ int notabs(VOID_A)
 
 int keyflush(VOID_A)
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	disable();
 	reg.x.ax = 0x0c00;
@@ -869,7 +878,7 @@ int reset;
 		cc_quit = cc.t_quitc;
 		cc_eof = cc.t_eofc;
 		cc_eol = cc.t_brkc;
-		if (cc_erase != 255) cc_erase = dupttyio.sg_erase;
+		if (cc_erase != K_UNDEF) cc_erase = dupttyio.sg_erase;
 #else
 # ifdef	VINTR
 		cc_intr = dupttyio.c_cc[VINTR];
@@ -883,7 +892,7 @@ int reset;
 # ifdef	VEOL
 		cc_eol = dupttyio.c_cc[VEOL];
 # endif
-		if (cc_erase != 255) cc_erase = dupttyio.c_cc[VERASE];
+		if (cc_erase != K_UNDEF) cc_erase = dupttyio.c_cc[VERASE];
 #endif
 #ifndef	USETERMINFO
 		ospeed = getspeed(dupttyio);
@@ -1051,8 +1060,11 @@ int notabs(VOID_A)
 int keyflush(VOID_A)
 {
 #ifdef	USESGTTY
-	int i = FREAD;
+	int i;
+#endif
 
+#ifdef	USESGTTY
+	i = FREAD;
 	VOID_C Xioctl(ttyio, TIOCFLUSH, &i);
 #else	/* !USESGTTY */
 # ifdef	USETERMIOS
@@ -1201,7 +1213,7 @@ int no;
 }
 
 static int NEAR err2(mes)
-char *mes;
+CONST char *mes;
 {
 	int duperrno;
 
@@ -1229,7 +1241,7 @@ char *mes;
 
 #if	!MSDOS
 static char *NEAR tstrdup(s)
-char *s;
+CONST char *s;
 {
 	char *cp;
 
@@ -1282,8 +1294,8 @@ static int NEAR defaultterm(VOID_A)
 #if	!MSDOS
 	for (i = 0; i <= K_MAX - K_MIN; i++) keyseq[i].code = K_MIN + i;
 	for (i = 21; i <= 30; i++)
-		keyseq[K_F(i) - K_MIN].code = K_F(i - 20) | 01000;
-#endif	/* !MSDOS */
+		keyseq[K_F(i) - K_MIN].code = K_F(i - 20) | K_ALTERNATE;
+#endif
 
 	return(0);
 }
@@ -1301,7 +1313,7 @@ int *yp, *xp;
 int getxy(xp, yp)
 int *xp, *yp;
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	reg.x.ax = 0x0300;
 	reg.h.bh = getbiosbyte(CURRPAGE);
@@ -1354,8 +1366,7 @@ int *xp, *yp;
 	format = SIZEFMT;
 	keyflush();
 # if	MSDOS
-	for (i = 0; i < strsize(GETSIZE); i++)
-		bdos(0x06, GETSIZE[i], 0);
+	for (i = 0; i < strsize(GETSIZE); i++) bdos(0x06, GETSIZE[i], 0);
 # else
 	if (!usegetcursor) return(-1);
 	tputs2(GETSIZE, 1);
@@ -1364,7 +1375,7 @@ int *xp, *yp;
 
 	buf[0] = '\0';
 	do {
-		if (!kbhit2(1000000L * 5)) break;
+		if (!kbhit2(WAITKEYPAD * 1000L * 10)) break;
 # if	MSDOS
 		buf[0] = bdos(0x07, 0x00, 0);
 # else
@@ -1409,7 +1420,7 @@ int *xp, *yp;
 
 #if	MSDOS
 char *tparamstr(s, arg1, arg2)
-char *s;
+CONST char *s;
 int arg1, arg2;
 {
 	char *cp;
@@ -1421,7 +1432,7 @@ int arg1, arg2;
 
 /*ARGSUSED*/
 int getterment(s)
-char *s;
+CONST char *s;
 {
 	if ((termflags & F_TERMENT) || (termflags & F_INPROGRESS)) return(-1);
 	termflags |= F_INPROGRESS;
@@ -1439,14 +1450,20 @@ char *s;
 #else	/* !MSDOS */
 
 char *tparamstr(s, arg1, arg2)
-char *s;
+CONST char *s;
 int arg1, arg2;
 {
+# ifndef	USETERMINFO
+	printbuf_t pbuf;
+	char *tmp;
+	int i, j, n, pop, err, args[2];
+# endif
+
 # ifdef	USETERMINFO
 #  ifdef	DEBUG
 	if (!s) return(NULL);
 	_mtrace_file = "tparm(start)";
-	s = tparm(s, arg1, arg2, 0, 0, 0, 0, 0, 0, 0);
+	s = tparm((char *)s, arg1, arg2, 0, 0, 0, 0, 0, 0, 0);
 	if (_mtrace_file) _mtrace_file = NULL;
 	else {
 		_mtrace_file = "tparm(end)";
@@ -1454,16 +1471,12 @@ int arg1, arg2;
 	}
 	if (!s) return(NULL);
 #  else
-	if (!s || !(s = tparm(s, arg1, arg2, 0, 0, 0, 0, 0, 0, 0)))
+	if (!s || !(s = tparm((char *)s, arg1, arg2, 0, 0, 0, 0, 0, 0, 0)))
 		return(NULL);
 #  endif
 
 	return(tstrdup(s));
 # else	/* !USETERMINFO */
-	printbuf_t pbuf;
-	char *tmp;
-	int i, j, n, pop, err, args[2];
-
 	if (!s || !(pbuf.buf = (char *)malloc(1))) return(NULL);
 	pbuf.ptr = pbuf.size = 0;
 	pbuf.flags = VF_NEW;
@@ -1560,12 +1573,14 @@ int arg1, arg2;
 
 static char *NEAR tgetstr2(term, s)
 char **term;
-char *s;
+CONST char *s;
 {
 # ifndef	USETERMINFO
 	char strbuf[TERMCAPSIZE];
 	char *p;
+# endif
 
+# ifndef	USETERMINFO
 	p = strbuf;
 #  ifdef	DEBUG
 	_mtrace_file = "tgetstr(start)";
@@ -1588,43 +1603,49 @@ char *s;
 }
 
 static char *NEAR tgetstr3(term, str1, str2)
-char **term, *str1, *str2;
+char **term;
+CONST char *str1, *str2;
 {
-# ifdef	USETERMINFO
-	if (str1) str1 = tstrdup(str1);
-	else str1 = tparamstr(str2, 1, 1);
-# else	/* !USETERMINFO */
+# ifndef	USETERMINFO
 	char *p, strbuf[TERMCAPSIZE];
+# endif
+	char *str;
 
+# ifdef	USETERMINFO
+	str = (str1) ? tstrdup(str1) : tparamstr(str2, 1, 1);
+# else	/* !USETERMINFO */
 	p = strbuf;
-
 #  ifdef	DEBUG
 	_mtrace_file = "tgetstr(start)";
-	str1 = tgetstr(str1, &p);
+	str = tgetstr(str1, &p);
 	if (_mtrace_file) _mtrace_file = NULL;
 	else {
 		_mtrace_file = "tgetstr(end)";
 		malloc(0);	/* dummy alloc */
 	}
-	if (str1) str1 = tstrdup(str1);
+	if (str) str = tstrdup(str);
 	else {
 		_mtrace_file = "tgetstr(start)";
-		str2 = tgetstr(str2, &p);
+		str = tgetstr(str2, &p);
 		if (_mtrace_file) _mtrace_file = NULL;
 		else {
 			_mtrace_file = "tgetstr(end)";
 			malloc(0);	/* dummy alloc */
 		}
-		str1 = tparamstr(str2, 1, 1);
+		str = tparamstr(str, 1, 1);
 	}
 #  else
-	if ((str1 = tgetstr(str1, &p))) str1 = tstrdup(str1);
-	else str1 = tparamstr(tgetstr(str2, &p), 1, 1);
+	str = tgetstr(str1, &p);
+	if (str) str = tstrdup(str);
+	else {
+		str = tgetstr(str2, &p);
+		str = tparamstr(str, 1, 1);
+	}
 #  endif
 # endif	/* !USETERMINFO */
-	if (str1) {
+	if (str) {
 		if (*term) free(*term);
-		*term = str1;
+		*term = str;
 	}
 
 	return(*term);
@@ -1632,7 +1653,7 @@ char **term, *str1, *str2;
 
 static char *NEAR tgetkeyseq(n, s)
 int n;
-char *s;
+CONST char *s;
 {
 	char *cp;
 	int i, j;
@@ -1640,20 +1661,22 @@ char *s;
 	n -= K_MIN;
 	cp = NULL;
 	if (!tgetstr2(&cp, s)) return(NULL);
-	if (keyseq[n].str) {
-		free(keyseq[n].str);
-		keyseq[n].str = NULL;
-	}
+	if ((keyseq[n].code & K_ALTERNATE) && *cp != '\033') return(NULL);
 	for (i = 0; i <= K_MAX - K_MIN; i++) {
-		if (!(keyseq[i].str)) continue;
+		if (i == n || !(keyseq[i].str)) continue;
 		for (j = 0; cp[j]; j++)
 			if ((cp[j] & 0x7f) != (keyseq[i].str[j] & 0x7f))
 				break;
-		if (!cp[j]) {
-			free(keyseq[i].str);
+		if (cp[j]) continue;
+		free(keyseq[i].str);
+		if (alternate(keyseq[n].code) != alternate(keyseq[i].code))
 			keyseq[i].str = NULL;
+		else {
+			keyseq[i].str = keyseq[n].str;
+			keyseq[n].str = NULL;
 		}
 	}
+	if (keyseq[n].str) free(keyseq[n].str);
 	keyseq[n].str = cp;
 
 	return(cp);
@@ -1748,16 +1771,16 @@ static int NEAR sortkeyseq(VOID_A)
 }
 
 int getterment(s)
-char *s;
+CONST char *s;
 {
 # ifdef	IRIX
-	/* for STUPID winterm entry */
-	int winterm;
+	int winterm;		/* for STUPID winterm entry */
 # endif
 # ifndef	USETERMINFO
 	char buf[TERMCAPSIZE];
 # endif
-	char *cp, *term;
+	CONST char *term;
+	char *cp;
 	int i, j, dumb, dupdumbterm;
 
 	if ((termflags & F_TERMENT) || (termflags & F_INPROGRESS)) return(-1);
@@ -1765,7 +1788,7 @@ char *s;
 
 	dupdumbterm = dumbterm;
 	dumbterm = dumb = 0;
-	term = (s) ? s : (char *)getenv("TERM");
+	term = (s) ? s : (char *)getenv(ENVTERM);
 	if (!term || !*term) {
 		dumbterm = 1;
 		dumb = DUMBLISTSIZE;
@@ -1784,14 +1807,14 @@ char *s;
 # ifdef	USETERMINFO
 #  ifdef	DEBUG
 		_mtrace_file = "setupterm(start)";
-		setupterm(term, fileno(ttyout), &i);
+		setupterm((char *)term, fileno(ttyout), &i);
 		if (_mtrace_file) _mtrace_file = NULL;
 		else {
 			_mtrace_file = "setupterm(end)";
 			malloc(0);	/* dummy alloc */
 		}
 #  else
-		setupterm(term, fileno(ttyout), &i);
+		setupterm((char *)term, fileno(ttyout), &i);
 #  endif
 		if (i == 1) break;
 # else	/* !USETERMINFO */
@@ -1824,7 +1847,7 @@ char *s;
 	}
 
 	if (dupdumbterm >= 2 || !strcmp(term, "emacs")
-	|| ((cp = getenv("EMACS")) && !strcmp(cp, "t")))
+	|| ((cp = getenv(ENVEMACS)) && !strcmp(cp, "t")))
 		dumbterm = 3;
 
 	defaultterm();
@@ -1855,8 +1878,8 @@ char *s;
 		*(termstr[T_KEYPAD]) = *(termstr[T_NOKEYPAD]) = '\0';
 		for (i = K_DOWN; i <= K_RIGHT; i++) {
 			if (keyseq[i - K_MIN].str) {
-				 free(keyseq[i - K_MIN].str);
-				 keyseq[i - K_MIN].str = NULL;
+				free(keyseq[i - K_MIN].str);
+				keyseq[i - K_MIN].str = NULL;
 			}
 		}
 	}
@@ -1944,16 +1967,16 @@ char *s;
 	tgetkeyseq(K_LEFT, TERM_kl);
 	tgetkeyseq(K_HOME, TERM_kh);
 	tgetkeyseq(K_BS, TERM_kb);
-	tgetkeyseq(K_F(1), TERM_l1);
-	tgetkeyseq(K_F(2), TERM_l2);
-	tgetkeyseq(K_F(3), TERM_l3);
-	tgetkeyseq(K_F(4), TERM_l4);
-	tgetkeyseq(K_F(5), TERM_l5);
-	tgetkeyseq(K_F(6), TERM_l6);
-	tgetkeyseq(K_F(7), TERM_l7);
-	tgetkeyseq(K_F(8), TERM_l8);
-	tgetkeyseq(K_F(9), TERM_l9);
-	tgetkeyseq(K_F(10), TERM_la);
+	tgetkeyseq(K_F(1), TERM_k1);
+	tgetkeyseq(K_F(2), TERM_k2);
+	tgetkeyseq(K_F(3), TERM_k3);
+	tgetkeyseq(K_F(4), TERM_k4);
+	tgetkeyseq(K_F(5), TERM_k5);
+	tgetkeyseq(K_F(6), TERM_k6);
+	tgetkeyseq(K_F(7), TERM_k7);
+	tgetkeyseq(K_F(8), TERM_k8);
+	tgetkeyseq(K_F(9), TERM_k9);
+	tgetkeyseq(K_F(10), TERM_k10);
 	tgetkeyseq(K_F(11), TERM_F1);
 	tgetkeyseq(K_F(12), TERM_F2);
 	tgetkeyseq(K_F(13), TERM_F3);
@@ -1964,17 +1987,16 @@ char *s;
 	tgetkeyseq(K_F(18), TERM_F8);
 	tgetkeyseq(K_F(19), TERM_F9);
 	tgetkeyseq(K_F(20), TERM_FA);
-	tgetkeyseq(K_F(21), TERM_k1);
-	tgetkeyseq(K_F(22), TERM_k2);
-	tgetkeyseq(K_F(23), TERM_k3);
-	tgetkeyseq(K_F(24), TERM_k4);
-	tgetkeyseq(K_F(25), TERM_k5);
-	tgetkeyseq(K_F(26), TERM_k6);
-	tgetkeyseq(K_F(27), TERM_k7);
-	tgetkeyseq(K_F(28), TERM_k8);
-	tgetkeyseq(K_F(29), TERM_k9);
-	tgetkeyseq(K_F(30), TERM_k10);
-	tgetkeyseq(K_F(30), TERM_k0);
+	tgetkeyseq(K_F(21), TERM_l1);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(22), TERM_l2);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(23), TERM_l3);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(24), TERM_l4);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(25), TERM_l5);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(26), TERM_l6);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(27), TERM_l7);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(28), TERM_l8);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(29), TERM_l9);	/* Hack for NEWS-OS bug */
+	tgetkeyseq(K_F(30), TERM_la);	/* Hack for NEWS-OS bug */
 	tgetkeyseq(K_DL, TERM_kL);
 	tgetkeyseq(K_IL, TERM_kA);
 	tgetkeyseq(K_DC, TERM_kD);
@@ -2002,9 +2024,8 @@ char *s;
 			for (j = 0; keyseq[i].str[j]; j++)
 				keyseq[i].str[j] &= 0x7f;
 			keyseq[i].len = j;
-			if (cc_erase != 255 && j == 1
-			&& keyseq[i].str[0] == cc_erase)
-				cc_erase = 255;
+			if (j == 1 && keyseq[i].str[0] == cc_erase)
+				cc_erase = K_UNDEF;
 		}
 	}
 	sortkeyseq();
@@ -2038,12 +2059,28 @@ int freeterment(VOID_A)
 	return(0);
 }
 
+int regetterment(s, tty)
+CONST char *s;
+int tty;
+{
+	int n, wastty;
+
+	if ((wastty = isttyiomode) && tty) stdiomode();
+	n = freeterment();
+	if (getterment(s) < 0) n = -1;
+	if (wastty && tty) ttyiomode(wastty - 1);
+
+	return(n);
+}
+
 int setdefterment(VOID_A)
 {
-	freeterment();
+	int n;
+
+	n = freeterment();
 	defaultterm();
 
-	return(0);
+	return(n);
 }
 
 int setdefkeyseq(VOID_A)
@@ -2063,7 +2100,7 @@ int setdefkeyseq(VOID_A)
 	}
 	for (i = 0; i <= K_MAX - K_MIN; i++) keyseq[i].code = K_MIN + i;
 	for (i = 21; i <= 30; i++)
-		keyseq[K_F(i) - K_MIN].code = K_F(i - 20) | 01000;
+		keyseq[K_F(i) - K_MIN].code = K_F(i - 20) | K_ALTERNATE;
 	sortkeyseq();
 
 	return(0);
@@ -2072,12 +2109,12 @@ int setdefkeyseq(VOID_A)
 int getdefkeyseq(kp)
 keyseq_t *kp;
 {
-	char *cp;
+	CONST char *cp;
 
 	if (kp -> code < K_MIN || kp -> code > K_MAX) /*EMPTY*/;
 	else if (!(cp = defkeyseq[kp -> code - K_MIN])) /*EMPTY*/;
 	else {
-		kp -> str = cp;
+		kp -> str = (char *)cp;
 		kp -> len = strlen(cp);
 		return(0);
 	}
@@ -2106,7 +2143,7 @@ int len;
 
 	if (str) {
 		for (i = 0; i <= K_MAX - K_MIN; i++) {
-			if ((keyseq[i].code & 0777) == n
+			if (alternate(keyseq[i].code) == n
 			|| !(keyseq[i].str) || keyseq[i].len != len)
 				continue;
 			if (!memcmp(str, keyseq[i].str, len)) {
@@ -2115,8 +2152,7 @@ int len;
 				keyseq[i].len = 0;
 			}
 		}
-		if (cc_erase != 255 && str[0] == cc_erase && !(str[1]))
-			cc_erase = 255;
+		if (str[0] == cc_erase && !(str[1])) cc_erase = K_UNDEF;
 	}
 	sortkeyseq();
 
@@ -2198,7 +2234,10 @@ keyseq_t *list;
 
 int initterm(VOID_A)
 {
-	if (!(termflags & F_TERMENT)) getterment(NULL);
+	int n;
+
+	n = 0;
+	if (!(termflags & F_TERMENT)) n = getterment(NULL);
 	termmode(1);
 	if (!dumbterm) {
 		putterm(T_KEYPAD);
@@ -2206,7 +2245,7 @@ int initterm(VOID_A)
 	}
 	termflags |= F_INITTERM;
 
-	return(0);
+	return(n);
 }
 
 int endterm(VOID_A)
@@ -2236,7 +2275,7 @@ int n;
 static int NEAR bioslocate(x, y)
 int x, y;
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	reg.x.ax = 0x0200;
 	reg.h.bh = getbiosbyte(CURRPAGE);
@@ -2250,7 +2289,7 @@ int x, y;
 static int NEAR biosscroll(d, sx, sy, ex, ey)
 int d, sx, sy, ex, ey;
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	if (sx > ex || sy > ey) return(0);
 	if (d >= 0) {
@@ -2274,7 +2313,7 @@ int d, sx, sy, ex, ey;
 static int NEAR biosputch(c, n)
 int c, n;
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	reg.h.ah = 0x09;
 	reg.h.al = (c & 0xff);
@@ -2289,7 +2328,7 @@ int c, n;
 static int NEAR bioscurstype(n)
 int n;
 {
-	__dpmi_regs reg;
+	__term_regs reg;
 
 	reg.x.ax = 0x0300;
 	reg.h.bh = getbiosbyte(CURRPAGE);
@@ -2434,7 +2473,7 @@ int n;
 }
 
 static int NEAR evalcsi(s)
-char *s;
+CONST char *s;
 {
 	static int savex, savey;
 	int i, x, y, w, h, n1, n2;
@@ -2571,7 +2610,7 @@ char *s;
 }
 
 int cputs2(s)
-char *s;
+CONST char *s;
 {
 	int i, n;
 
@@ -2593,7 +2632,7 @@ int c;
 }
 
 int cputs2(s)
-char *s;
+CONST char *s;
 {
 	int i, x, y;
 
@@ -2618,18 +2657,20 @@ char *s;
 int kbhit2(usec)
 long usec;
 {
-#if	defined (NOTUSEBIOS) || !defined (DJGPP) || (DJGPP < 2)
+#if	defined (NOTUSEBIOS) || defined (NOSELECT)
 	union REGS reg;
-#else	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
+#else	/* !NOTUSEBIOS && !NOSELECT */
 # ifndef	PC98
 	struct timeval tv;
 	int n;
 # endif
-#endif	/* !NOTUSEBIOS && DJGPP && DJGPP >= 2 */
+#endif	/* !NOTUSEBIOS && !NOSELECT */
 
+#if	defined (PC98) || defined (NOTUSEBIOS)
+	if (nextchar) return(1);
+#endif
 	if (ungetnum > 0) return(1);
 #ifdef	NOTUSEBIOS
-	if (nextchar) return(1);
 	reg.x.ax = 0x4406;
 	reg.x.bx = ttyio;
 	putterm(T_METAMODE);
@@ -2639,24 +2680,24 @@ long usec;
 	return((reg.x.flags & 1) ? 0 : reg.h.al);
 #else	/* !NOTUSEBIOS */
 # ifdef	PC98
-	if (nextchar) return(1);
 	reg.h.ah = 0x01;
 	int86(0x18, &reg, &reg);
 
 	return(reg.h.bh != 0);
 # else	/* !PC98 */
-#  if	defined (DJGPP) && (DJGPP >= 2)
-	tv.tv_sec = (time_t)usec / (time_t)1000000;
-	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
-
-	return(n);
-#  else	/* !DJGPP || DJGPP < 2 */
+#  ifdef	NOSELECT
 	reg.h.ah = 0x01;
 	int86(0x16, &reg, &reg);
 
 	return((reg.x.flags & 0x40) ? 0 : 1);
-#  endif	/* !DJGPP || DJGPP < 2 */
+#  else	/* !NOSELECT */
+	tv.tv_sec = (time_t)usec / (time_t)1000000;
+	tv.tv_usec = (time_t)usec % (time_t)1000000;
+	if ((n = sureselect(1, &ttyio, NULL, &tv, SEL_TTYIO)) < 0)
+		err2("select()");
+
+	return(n);
+#  endif	/* !NOSELECT */
 # endif	/* !PC98 */
 #endif	/* !NOTUSEBIOS */
 }
@@ -2671,7 +2712,9 @@ int getch2(VOID_A)
 #endif
 #if	defined (PC98) || defined (NOTUSEBIOS)
 	int ch;
+#endif
 
+#if	defined (PC98) || defined (NOTUSEBIOS)
 	if (nextchar) {
 		ch = nextchar;
 		nextchar = '\0';
@@ -2750,8 +2793,8 @@ u_char tbuf[];
 #endif
 
 /*ARGSUSED*/
-int getkey2(sig)
-int sig;
+int getkey2(sig, code)
+int sig, code;
 {
 #if	!defined (DJGPP) || defined (NOTUSEBIOS) || defined (PC98)
 	static u_char tbuf1[3] = {0xff, 0xff, 0xff};
@@ -2765,7 +2808,7 @@ int sig;
 #if	defined (DJGPP) && !defined (NOTUSEBIOS) && !defined (PC98)
 	do {
 		i = kbhit2(1000000L / SENSEPERSEC);
-# if	defined (DJGPP) && (DJGPP >= 2)
+# ifndef	NOSELECT
 		if (sig && !(--count)) {
 			count = SENSEPERSEC;
 			raise(sig);
@@ -2799,8 +2842,7 @@ int sig;
 		default:
 			break;
 	}
-#if	defined (PC98) || defined (NOTUSEBIOS) \
-|| (defined (DJGPP) && (DJGPP >= 2))
+#if	defined (PC98) || defined (NOTUSEBIOS) || !defined (NOSELECT)
 	else if (!kbhit2(WAITKEYPAD * 1000L)) /*EMPTY*/;
 #endif
 	else if ((ch = getch2()) == EOF) ch = K_NOKEY;
@@ -2815,13 +2857,12 @@ int sig;
 	return(ch);
 }
 
-int ungetch2(c)
+int ungetkey2(c)
 int c;
 {
-	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
-	if (ungetnum)
-		memmove((char *)&(ungetbuf[1]), (char *)&(ungetbuf[0]),
-			ungetnum * sizeof(u_char));
+	if (ungetnum >= arraysize(ungetbuf)) return(EOF);
+	memmove((char *)&(ungetbuf[1]), (char *)&(ungetbuf[0]),
+		ungetnum * sizeof(u_char));
 	ungetbuf[0] = c;
 	ungetnum++;
 
@@ -2886,7 +2927,7 @@ int c;
 }
 
 int cputs2(s)
-char *s;
+CONST char *s;
 {
 	if (!s) return(0);
 
@@ -2900,7 +2941,7 @@ tputs_t c;
 }
 
 int tputs2(s, n)
-char *s;
+CONST char *s;
 int n;
 {
 	return(tputs(s, n, putch3));
@@ -2915,18 +2956,28 @@ int n;
 	return(tputs2(termstr[n], n_line));
 }
 
+VOID checksuspend(VOID_A)
+{
+	if (suspended) ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
+	suspended = 0;
+}
+
 int kbhit2(usec)
 long usec;
 {
+# ifndef	NOSELECT
+	struct timeval tv;
+	int n;
+# endif
+
+	if (ungetnum > 0) return(1);
 # ifdef	NOSELECT
 	return((usec) ? 1 : 0);
 # else
-	struct timeval tv;
-	int n;
-
 	tv.tv_sec = (time_t)usec / (time_t)1000000;
 	tv.tv_usec = (time_t)usec % (time_t)1000000;
-	if ((n = readselect(1, &ttyio, NULL, &tv)) < 0) err2("select()");
+	if ((n = sureselect(1, &ttyio, NULL, &tv, SEL_TTYIO)) < 0)
+		err2("select()");
 
 	return(n);
 # endif
@@ -2935,21 +2986,22 @@ long usec;
 int getch2(VOID_A)
 {
 	u_char ch;
-	int i;
+	int n;
 
-	do {
-		if (suspended) {
-			ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
-			suspended = 0;
-		}
-	} while ((i = read(ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
-	if (i < (int)sizeof(u_char)) return(EOF);
+	if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
+	for (;;) {
+		checksuspend();
+		n = read(ttyio, &ch, sizeof(u_char));
+		if (n >= 0 || errno != EINTR) break;
+	}
+	if (n < (int)sizeof(u_char)) return(EOF);
 
 	return((int)ch);
 }
 
-int getkey2(sig)
-int sig;
+/*ARGSUSED*/
+int getkey2(sig, code)
+int sig, code;
 {
 	static int count = SENSEPERSEC;
 	kstree_t *p;
@@ -2961,14 +3013,7 @@ int sig;
 			count = SENSEPERSEC;
 			kill(getpid(), sig);
 		}
-		if (suspended) {
-			ttyiomode((isttyiomode) ? isttyiomode - 1 : 0);
-			suspended = 0;
-		}
 		if (keywaitfunc && (ch = (*keywaitfunc)()) < 0) return(ch);
-# ifndef	TIOCSTI
-		if (ungetnum > 0) return((int)ungetbuf[--ungetnum]);
-# endif
 	} while (!key);
 
 	if ((key = ch = getch2()) == EOF) return(K_NOKEY);
@@ -2976,9 +3021,9 @@ int sig;
 	else if (key != C_EKANA) /*EMPTY*/;
 #  if	!defined (_NOKANJICONV)
 #   ifdef	CODEEUC
-	else if (inputkcode != EUC && inputkcode != NOCNV) /*EMPTY*/;
+	else if (code != EUC && code != NOCNV) /*EMPTY*/;
 #   else
-	else if (inputkcode != EUC) /*EMPTY*/;
+	else if (code != EUC) /*EMPTY*/;
 #   endif
 #  endif	/* !_NOKANJICONV */
 	else {
@@ -2990,7 +3035,7 @@ int sig;
 	}
 # endif	/* !_NOKANJICONV || CODEEUC */
 
-	if (cc_erase != 255 && key == cc_erase) return(K_BS);
+	if (cc_erase != K_UNDEF && key == cc_erase) return(K_BS);
 	if (!(p = keyseqtree)) return(key);
 
 	if (key == K_ESC) {
@@ -3012,8 +3057,7 @@ int sig;
 			if (key == keyseq[p -> next[j].key].str[0]) break;
 		if (j >= p -> num) return(key);
 		p = &(p -> next[j]);
-		if (keyseq[p -> key].len == 1)
-			return(keyseq[p -> key].code);
+		if (keyseq[p -> key].len == 1) return(keyseq[p -> key].code);
 		if (!kbhit2(WAITKEYPAD * 1000L) || (ch = getch2()) == EOF)
 			return(key);
 	}
@@ -3034,38 +3078,48 @@ int sig;
 		ungetch2(keyseq[p -> key].str[j]);
 	}
 	ungetch2(ch);
+
 	return(key);
 }
 
-int getkey3(sig)
-int sig;
+int getkey3(sig, code)
+int sig, code;
 {
 	int ch;
 
-	if ((ch = getkey2(sig)) < 0) return(ch);
+	if ((ch = getkey2(sig, code)) < 0) return(ch);
 	if (ch >= K_F('*') && ch < K_DL) {
 		if (ch == K_F('?')) ch = K_CR;
 		else ch -= K_F0;
 	}
 
-	return(ch & 01777);
+	return(alternate(ch));
 }
 
-int ungetch2(c)
+static int NEAR ungetch2(c)
+int c;
+{
+	if (ungetnum >= arraysize(ungetbuf)) return(EOF);
+	memmove((char *)&(ungetbuf[1]), (char *)&(ungetbuf[0]),
+		ungetnum * sizeof(u_char));
+	ungetbuf[0] = c;
+	ungetnum++;
+
+	return(c);
+}
+
+int ungetkey2(c)
 int c;
 {
 # ifdef	TIOCSTI
 	u_char ch;
+# endif
 
+# ifdef	TIOCSTI
 	ch = c;
 	Xioctl(ttyio, TIOCSTI, &ch);
 # else
-	if (ungetnum >= arraysize(ungetbuf) - 1) return(EOF);
-	if (ungetnum)
-		memmove((char *)&(ungetbuf[1]), (char *)&(ungetbuf[0]),
-			ungetnum * sizeof(u_char));
-	ungetbuf[0] = c;
-	ungetnum++;
+	if (ungetch2(c) == EOF) return(EOF);
 # endif
 
 	return(c);
@@ -3169,11 +3223,13 @@ int xmax, ymax;
 int setwsize(fd, xmax, ymax)
 int fd, xmax, ymax;
 {
+# ifndef	NOTERMWSIZE
+	termwsize_t ws;
+# endif
+
 # ifdef	NOTERMWSIZE
 	return(0);
 # else	/* !NOTERMWSIZE */
-	termwsize_t ws;
-
 	memset((char *)&ws, 0, sizeof(ws));
 	VOID_C Xioctl(fd, REQGETWS, &ws);
 
@@ -3236,7 +3292,7 @@ int cputnl(VOID_A)
 }
 
 int kanjiputs(s)
-char *s;
+CONST char *s;
 {
 	return(cprintf2("%k", s));
 }
