@@ -23,33 +23,15 @@
 #endif
 
 #ifdef	NOFILEMODE
-# ifdef	S_IRUSR
-# undef	S_IRUSR
-# endif
-# ifdef	S_IWUSR
-# undef	S_IWUSR
-# endif
-# ifdef	S_IXUSR
-# undef	S_IXUSR
-# endif
-# ifdef	S_IRGRP
-# undef	S_IRGRP
-# endif
-# ifdef	S_IWGRP
-# undef	S_IWGRP
-# endif
-# ifdef	S_IXGRP
-# undef	S_IXGRP
-# endif
-# ifdef	S_IROTH
-# undef	S_IROTH
-# endif
-# ifdef	S_IWOTH
-# undef	S_IWOTH
-# endif
-# ifdef	S_IXOTH
-# undef	S_IXOTH
-# endif
+#undef	S_IRUSR
+#undef	S_IWUSR
+#undef	S_IXUSR
+#undef	S_IRGRP
+#undef	S_IWGRP
+#undef	S_IXGRP
+#undef	S_IROTH
+#undef	S_IWOTH
+#undef	S_IXOTH
 #define	S_IRUSR	00400
 #define	S_IWUSR	00200
 #define	S_IXUSR	00100
@@ -143,7 +125,7 @@ typedef struct _namelist {
 
 typedef struct _strtable {
 	u_short no;
-	char *str;
+	CONST char *str;
 } strtable;
 
 typedef struct _lockbuf_t {
@@ -156,6 +138,7 @@ typedef struct _lockbuf_t {
 #define	LCK_FLOCK		0001
 #define	LCK_INVALID		0002
 #define	LCK_STREAM		0004
+#define	LCK_MAXRETRY		32
 
 typedef struct _bindtable {
 	short key;
@@ -164,13 +147,13 @@ typedef struct _bindtable {
 } bindtable;
 
 typedef struct _functable {
-	int (*func)__P_((char *));
-	char *ident;
+	int (*func)__P_((CONST char *));
+	CONST char *ident;
 #ifndef	_NOJPNMES
-	char *hmes;
+	CONST char *hmes;
 #endif
 #if	!defined (_NOENGMES) || defined (_NOJPNMES)
-	char *hmes_eng;
+	CONST char *hmes_eng;
 #endif
 	u_char status;
 } functable;
@@ -281,10 +264,13 @@ typedef struct _winvartable {
 
 extern winvartable winvar[];
 #ifdef	_NOSPLITWIN
-#define	win			0
 #define	windows			1
 #else
 extern int windows;
+#endif
+#if	defined (_NOSPLITWIN) && defined (_NOPTY)
+#define	win			0
+#else
 extern int win;
 #endif
 #ifndef	_NOARCHIVE
@@ -351,7 +337,7 @@ typedef struct _userfunctable {
 #endif	/* _NOORIGSHELL */
 
 typedef struct _builtintable {
-	int (NEAR *func)__P_((int, char *[]));
+	int (NEAR *func)__P_((int, char *CONST []));
 	char *ident;
 } builtintable;
 
@@ -422,6 +408,7 @@ typedef struct _builtintable {
 #define	TCH_CHANGE		00100
 #define	TCH_MASK		00200
 #define	TCH_MODEEXE		00400
+#define	TCH_IGNOREERR		01000
 
 #define	ATR_EXCLUSIVE		3
 #define	ATR_MODEONLY		1
