@@ -7,8 +7,10 @@
 #include "machine.h"
 #include <stdio.h>
 #include <fcntl.h>
-#include <io.h>
 
+#if	MSDOS
+#include <io.h>
+#endif
 #ifndef	NOUNISTDH
 #include <unistd.h>
 #endif
@@ -37,33 +39,39 @@
 #define	int21call(reg, sreg)	intcall(0x21, reg, sreg)
 
 #ifdef	FD
-extern int _dospath __P_((char *));
+extern int _dospath __P_((CONST char *));
 extern char *gendospath __P_((char *, int, int));
-extern char *strdelim __P_((char *, int));
-extern char *strrdelim __P_((char *, int));
-extern int isdelim __P_((char *, int));
+extern char *strdelim __P_((CONST char *, int));
+extern char *strrdelim __P_((CONST char *, int));
+extern int isdelim __P_((CONST char *, int));
 extern char *strcatdelim __P_((char *));
-extern char *strcatdelim2 __P_((char *, char *, char *));
-extern int isdotdir __P_((char *));
-extern char curpath[];
+extern char *strcatdelim2 __P_((char *, CONST char *, CONST char *));
+extern int isdotdir __P_((CONST char *));
+# ifdef	DOUBLESLASH
+extern int isdslash __P_((CONST char *));
+# endif
+extern CONST char curpath[];
 #else
-static int NEAR _dospath __P_((char *));
+static int NEAR _dospath __P_((CONST char *));
 static char *NEAR gendospath __P_((char *, int, int));
-static char *NEAR strdelim __P_((char *, int));
-static char *NEAR strrdelim __P_((char *, int));
-static int NEAR isdelim __P_((char *, int));
+static char *NEAR strdelim __P_((CONST char *, int));
+static char *NEAR strrdelim __P_((CONST char *, int));
+static int NEAR isdelim __P_((CONST char *, int));
 static char *NEAR strcatdelim __P_((char *));
-static char *NEAR strcatdelim2 __P_((char *, char *, char *));
-static int NEAR isdotdir __P_((char *));
-static char curpath[] = ".";
+static char *NEAR strcatdelim2 __P_((char *, CONST char *, CONST char *));
+static int NEAR isdotdir __P_((CONST char *));
+# ifdef	DOUBLESLASH
+static int isdslash __P_((CONST char *));
+# endif
+static CONST char curpath[] = ".";
 #endif
 
 #ifdef	DJGPP
-static int NEAR dos_putpath __P_((char *, int));
+static int NEAR dos_putpath __P_((CONST char *, int));
 #endif
-static char *NEAR duplpath __P_((char *));
+static char *NEAR duplpath __P_((CONST char *));
 #ifndef	_NODOSDRIVE
-static char *NEAR regpath __P_((char *, char *));
+static char *NEAR regpath __P_((CONST char *, char *));
 static VOID NEAR biosreset __P_((int));
 static int NEAR _biosdiskio __P_((int, int, int, int,
 		u_char *, int, int, int));
@@ -78,17 +86,17 @@ static int NEAR lockdrive __P_((int));
 static int NEAR unlockdrive __P_((int, int));
 static int NEAR dosrawdiskio __P_((int, u_long, u_char *, int, int, int));
 #endif
-static int NEAR dos_findfirst __P_((char *, u_int, struct dosfind_t *));
+static int NEAR dos_findfirst __P_((CONST char *, u_int, struct dosfind_t *));
 static int NEAR dos_findnext __P_((struct dosfind_t *));
 #ifndef	_NOUSELFN
-static int NEAR lfn_findfirst __P_((u_int *, char *,
+static int NEAR lfn_findfirst __P_((u_int *, CONST char *,
 		u_int, struct lfnfind_t *));
 static int NEAR lfn_findnext __P_((u_int, struct lfnfind_t *));
 static int NEAR lfn_findclose __P_((u_int));
 #endif
 #ifndef	_NOUSELFN
 static int NEAR gendosname __P_((char *));
-static int NEAR unixrenamedir __P_((char *, char *));
+static int NEAR unixrenamedir __P_((CONST char *, CONST char *));
 #endif
 static u_int NEAR getdosmode __P_((u_int));
 static u_int NEAR putdosmode __P_((u_int));
@@ -96,6 +104,8 @@ static time_t NEAR getdostime __P_((u_int, u_int));
 #ifndef	_NOUSELFN
 static int NEAR putdostime __P_((u_short *, u_short *, time_t));
 #endif
+static int NEAR alter_findfirst __P_((CONST char *, u_int *,
+		struct dosfind_t *, struct lfnfind_t *));
 
 #ifndef	_NODOSDRIVE
 int dosdrive = 0;
@@ -144,11 +154,14 @@ static u_char int25bin[] = {
 };
 static int (far *doint25)__P_((VOID_A)) = NULL;
 #endif	/* !DJGPP && !_NODOSDRIVE */
+#ifdef	DOUBLESLASH
+static int dsdrive = 0;
+#endif
 
 
 #ifndef	FD
 static int NEAR _dospath(path)
-char *path;
+CONST char *path;
 {
 	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
 }
@@ -166,14 +179,14 @@ int drive, c;
 }
 
 static char *NEAR strdelim(s, d)
-char *s;
+CONST char *s;
 int d;
 {
 	int i;
 
-	if (d && _dospath(s)) return(s + 1);
+	if (d && _dospath(s)) return((char *)&(s[1]));
 	for (i = 0; s[i]; i++) {
-		if (s[i] == _SC_) return(&(s[i]));
+		if (s[i] == _SC_) return((char *)&(s[i]));
 # ifdef	BSPATHDELIM
 		if (iskanji1(s, i)) i++;
 # endif
@@ -183,13 +196,13 @@ int d;
 }
 
 char *strrdelim(s, d)
-char *s;
+CONST char *s;
 int d;
 {
-	char *cp;
+	CONST char *cp;
 	int i;
 
-	if (d && _dospath(s)) cp = s + 1;
+	if (d && _dospath(s)) cp = &(s[1]);
 	else cp = NULL;
 	for (i = 0; s[i]; i++) {
 		if (s[i] == _SC_) cp = &(s[i]);
@@ -198,11 +211,11 @@ int d;
 # endif
 	}
 
-	return(cp);
+	return((char *)cp);
 }
 
 static int NEAR isdelim(s, ptr)
-char *s;
+CONST char *s;
 int ptr;
 {
 # ifdef	BSPATHDELIM
@@ -256,7 +269,8 @@ char *s;
 }
 
 static char *NEAR strcatdelim2(buf, s1, s2)
-char *buf, *s1, *s2;
+char *buf;
+CONST char *s1, *s2;
 {
 	char *cp;
 	int i, len;
@@ -307,7 +321,7 @@ char *buf, *s1, *s2;
 }
 
 int isdotdir(s)
-char *s;
+CONST char *s;
 {
 	if (s[0] != '.') /*EMPTY*/;
 	else if (!s[1]) return(2);
@@ -316,11 +330,24 @@ char *s;
 
 	return(0);
 }
+
+# ifdef	DOUBLESLASH
+static int isdslash(s)
+CONST char *s;
+{
+	char *cp;
+
+	if (s[0] != _SC_ || s[1] != _SC_ || !s[2] || s[2] == _SC_) return(0);
+	if ((cp = strdelim(&(s[2]), 0))) return(cp - s);
+
+	return(strlen(s));
+}
+# endif	/* DOUBLESLASH */
 #endif	/* !FD */
 
 #ifdef	DJGPP
 static int NEAR dos_putpath(path, offset)
-char *path;
+CONST char *path;
 int offset;
 {
 	int i;
@@ -333,17 +360,27 @@ int offset;
 #endif
 
 static char *NEAR duplpath(path)
-char *path;
+CONST char *path;
 {
+#ifdef	DOUBLESLASH
+	int len;
+#endif
 	static char buf[MAXPATHLEN];
 	int i, j, ps;
 
-	if (path == buf) return(path);
+	if (path == buf) return((char *)path);
 	i = j = ps = 0;
 	if (_dospath(path)) {
 		buf[j++] = path[i++];
 		buf[j++] = path[i++];
 	}
+#ifdef	DOUBLESLASH
+	if ((len = isdslash(path))) {
+		memcpy(buf, path, len);
+		i = j = len;
+	}
+	else
+#endif
 	if (path[i] == _SC_) buf[j++] = path[i++];
 	for (; path[i]; i++) {
 		if (path[i] == _SC_) ps = 1;
@@ -365,6 +402,9 @@ int getcurdrv(VOID_A)
 {
 	int drive;
 
+#ifdef	DOUBLESLASH
+	if (dsdrive) return(dsdrive);
+#endif
 	drive = (dos7access & D7_CAPITAL) ? 'A' : 'a';
 #ifndef	_NODOSDRIVE
 	if (lastdrive >= 0) drive += toupper2(lastdrive) - 'A';
@@ -383,9 +423,17 @@ int drive, nodir;
 #endif
 	struct SREGS sreg;
 	__dpmi_regs reg;
-	char *path;
+	CONST char *path;
 	int drv, olddrv;
 
+#ifdef	DOUBLESLASH
+	if (drive == '_') {
+		dsdrive = drive;
+		return(0);
+	}
+	dsdrive = '\0';
+#endif
+
 	errno = EINVAL;
 	if (islower2(drive)) {
 		drv = drive - 'a';
@@ -408,7 +456,7 @@ int drive, nodir;
 	if ((bdos(0x0e, drv, 0) & 0xff) < drv
 	|| (drive = (bdos(0x19, 0, 0) & 0xff)) != drv) {
 		bdos(0x0e, olddrv, 0);
-		seterrno(0x0f);		/* Invarid drive */
+		dosseterrno(0x0f);	/* Invarid drive */
 		return(-1);
 	}
 
@@ -460,6 +508,7 @@ int getdosver(VOID_A)
 }
 
 /*
+ *	 3: enable LFN access, but shared folder
  *	 2: enable LFN access, but FAT32
  *	 1: enable LFN access
  *	 0: do SFN access
@@ -470,13 +519,19 @@ int getdosver(VOID_A)
  */
 
 int supportLFN(path)
-char *path;
+CONST char *path;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
 	char drive, drv[4], buf[128];
 
+#ifdef	DOUBLESLASH
+	if (isdslash(path)) return(3);
+#endif
 	if (!path || !(drive = _dospath(path))) drive = getcurdrv();
+#ifdef	DOUBLESLASH
+	if (drive == '_') return(3);
+#endif
 #ifndef	_NODOSDRIVE
 	if (checkdrive(toupper2(drive) - 'A') > 0) return(-2);
 #endif
@@ -544,14 +599,15 @@ int drive;
 #ifndef	_NOUSELFN
 # ifndef	_NODOSDRIVE
 static char *NEAR regpath(path, buf)
-char *path, *buf;
+CONST char *path;
+char *buf;
 {
-	char *cp;
+	CONST char *cp;
 	int drive;
 
 	cp = path;
 	if (!(drive = _dospath(path))) drive = getcurdrv();
-	else if (*(cp += 2)) return(path);
+	else if (*(cp += 2)) return((char *)path);
 	strcpy(gendospath(buf, drive, (*cp) ? '\0' : '.'), cp);
 
 	return(buf);
@@ -559,7 +615,8 @@ char *path, *buf;
 # endif	/* !_NODOSDRIVE */
 
 char *shortname(path, alias)
-char *path, *alias;
+CONST char *path;
+char *alias;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
@@ -577,7 +634,7 @@ char *path, *alias;
 			else if (!dependdosfunc) return(NULL);
 		}
 # endif	/* !_NODOSDRIVE */
-		return(path);
+		return((char *)path);
 	}
 	reg.x.ax = 0x7160;
 	reg.x.cx = 0x8001;
@@ -595,6 +652,9 @@ char *path, *alias;
 # ifndef	BSPATHDELIM
 	adjustpname(alias);
 # endif
+# ifdef	DOUBLESLASH
+	if (isdslash(alias)) return(alias);
+# endif
 
 	if (!path || !(i = _dospath(path))) i = getcurdrv();
 	if (islower2(i)) *alias = tolower2(*alias);
@@ -604,7 +664,8 @@ char *path, *alias;
 #endif	/* !_NOUSELFN */
 
 char *unixrealpath(path, resolved)
-char *path, *resolved;
+CONST char *path;
+char *resolved;
 {
 #ifndef	_NODOSDRIVE
 	char buf[MAXPATHLEN];
@@ -619,6 +680,7 @@ char *path, *resolved;
 	reg.x.cx = 0;
 #else	/* !_NOUSELFN */
 	switch (supportLFN(path)) {
+		case 3:
 		case 2:
 		case 1:
 			reg.x.ax = 0x7160;
@@ -660,6 +722,9 @@ char *path, *resolved;
 #ifndef	BSPATHDELIM
 	adjustpname(resolved);
 #endif
+#ifdef	DOUBLESLASH
+	if (isdslash(resolved)) return(resolved);
+#endif
 
 	if (!path || !(i = _dospath(path))) i = getcurdrv();
 	else path += 2;
@@ -733,21 +798,22 @@ char *path;
 
 #ifndef	_NOUSELFN
 char *preparefile(path, alias)
-char *path, *alias;
+CONST char *path;
+char *alias;
 {
 	int i;
 
 	path = duplpath(path);
 	i = supportLFN(path);
 #ifdef	_NODOSDRIVE
-	if (i < 0) return(path);
+	if (i < 0) return((char *)path);
 #else	/* !_NODOSDRIVE */
-	if (i < -2) return(path);
+	if (i < -2) return((char *)path);
 #endif	/* !_NODOSDRIVE */
 
 	if ((alias = shortname(path, alias))) return(alias);
 #ifndef	_NODOSDRIVE
-	else if (i < 0 && errno == EACCES) return(path);
+	else if (i < 0 && errno == EACCES) return((char *)path);
 #endif
 
 	return(NULL);
@@ -1532,7 +1598,7 @@ int n, size, iswrite;
 #endif	/* !_NOUSELFN */
 
 static int NEAR dos_findfirst(path, attr, result)
-char *path;
+CONST char *path;
 u_int attr;
 struct dosfind_t *result;
 {
@@ -1552,7 +1618,10 @@ struct dosfind_t *result;
 #endif
 	sreg.ds = PTR_SEG(path);
 	reg.x.dx = PTR_OFF(path, sizeof(struct dosfind_t));
-	if (int21call(&reg, &sreg) < 0) return(-1);
+	if (int21call(&reg, &sreg) < 0) {
+		if (!errno) errno = ENOENT;
+		return(-1);
+	}
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct dosfind_t), result);
 #endif
@@ -1586,7 +1655,7 @@ struct dosfind_t *result;
 #ifndef	_NOUSELFN
 static int NEAR lfn_findfirst(fdp, path, attr, result)
 u_int *fdp;
-char *path;
+CONST char *path;
 u_int attr;
 struct lfnfind_t *result;
 {
@@ -1604,7 +1673,10 @@ struct lfnfind_t *result;
 	reg.x.dx = PTR_OFF(path, sizeof(struct lfnfind_t));
 	sreg.es = PTR_SEG(result);
 	reg.x.di = PTR_OFF(result, 0);
-	if (int21call(&reg, &sreg) < 0) return(-1);
+	if (int21call(&reg, &sreg) < 0) {
+		if (!errno) errno = ENOENT;
+		return(-1);
+	}
 	*fdp = reg.x.ax;
 #ifdef	DJGPP
 	dosmemget(__tb, sizeof(struct lfnfind_t), result);
@@ -1647,7 +1719,7 @@ u_int fd;
 #endif	/* !_NOUSELFN */
 
 DIR *unixopendir(dir)
-char *dir;
+CONST char *dir;
 {
 	DIR *dirp;
 	char *cp, path[MAXPATHLEN];
@@ -1849,7 +1921,7 @@ DIR *dirp;
 # ifndef	DJGPP
 /*ARGSUSED*/
 int unixmkdir(path, mode)
-char * path;
+CONST char * path;
 int mode;
 {
 	struct stat st;
@@ -1864,7 +1936,7 @@ int mode;
 # endif
 #else	/* !_NOUSELFN */
 int unixunlink(path)
-char *path;
+CONST char *path;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
@@ -1926,27 +1998,28 @@ char *s;
 }
 
 static int NEAR unixrenamedir(from, to)
-char *from, *to;
+CONST char *from, *to;
 {
 	DIR *dirp;
 	struct dirent *dp;
+	CONST char *cp;
 	char *fp, *tp, fbuf[MAXPATHLEN], tbuf[MAXPATHLEN];
 	int i;
 
 	strcpy(fbuf, to);
 	if (!(tp = strrdelim(fbuf, 1))) {
-		tp = to;
+		cp = to;
 		strcpy(fbuf, curpath);
 	}
 	else {
 		if (*tp == _SC_) tp++;
 		*tp = '\0';
-		tp = &(to[tp - fbuf]);
+		cp = &(to[tp - fbuf]);
 	}
 	if (!(to = unixrealpath(fbuf, tbuf))
 	|| !(from = unixrealpath(from, fbuf)))
 		return(-1);
-	strcpy(strcatdelim(tbuf), tp);
+	strcpy(strcatdelim(tbuf), cp);
 
 	for (i = 0; from[i]; i++) if (from[i] != to[i]) break;
 	if (!strdelim(&(from[i]), 0) && !strdelim(&(to[i]), 0)) return(-1);
@@ -1957,8 +2030,8 @@ char *from, *to;
 		return(-1);
 	}
 	i = strlen(from);
-	fp = strcatdelim(from);
-	tp = strcatdelim(to);
+	fp = strcatdelim(fbuf);
+	tp = strcatdelim(tbuf);
 	while ((dp = unixreaddir(dirp))) {
 		if (isdotdir(dp -> d_name)) continue;
 		strcpy(fp, dp -> d_name);
@@ -1969,14 +2042,14 @@ char *from, *to;
 		}
 	}
 	unixclosedir(dirp);
-	from[i] = '\0';
+	fbuf[i] = '\0';
 	if (unixrmdir(from) < 0) return(-1);
 
 	return(0);
 }
 
 int unixrename(from, to)
-char *from, *to;
+CONST char *from, *to;
 {
 	struct dosfind_t dbuf;
 	struct SREGS sreg;
@@ -2024,7 +2097,7 @@ char *from, *to;
 				break;
 
 			if (t > -3) {
-				if ((f = gendosname(to)) < 0) break;
+				if ((f = gendosname(tbuf)) < 0) break;
 				if (f > 0) {
 					ax = 0x7156;
 					continue;
@@ -2041,7 +2114,7 @@ char *from, *to;
 
 /*ARGSUSED*/
 int unixmkdir(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
 	struct SREGS sreg;
@@ -2072,7 +2145,7 @@ int mode;
 }
 
 int unixrmdir(path)
-char *path;
+CONST char *path;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
@@ -2098,7 +2171,7 @@ char *path;
 }
 
 int unixchdir(path)
-char *path;
+CONST char *path;
 {
 	struct SREGS sreg;
 	__dpmi_regs reg;
@@ -2132,6 +2205,7 @@ int size;
 #ifdef	DJGPP
 	char tmp[MAXPATHLEN];
 #endif
+	char *cp;
 	int drive;
 
 	drive = getcurdrv();
@@ -2141,9 +2215,15 @@ int size;
 	}
 	else
 #endif
-	if ((!pathname && !(pathname = (char *)malloc(size)))
-	|| !unixgetcurdir(gendospath(pathname, drive, _SC_), 0))
-		return(NULL);
+	if (!pathname && !(pathname = (char *)malloc(size))) return(NULL);
+	else {
+#ifdef	DOUBLESLASH
+		if (drive == '_') cp = pathname;
+		else
+#endif
+		cp = gendospath(pathname, drive, _SC_);
+		if (!unixgetcurdir(cp, 0)) return(NULL);
+	}
 
 	*pathname = (dos7access & D7_CAPITAL)
 		? toupper2(*pathname) : tolower2(*pathname);
@@ -2223,7 +2303,7 @@ time_t tim;
 #endif	/* !_NOUSELFN */
 
 int unixstatfs(path, buf)
-char *path;
+CONST char *path;
 statfs_t *buf;
 {
 #ifndef	_NOUSELFN
@@ -2256,7 +2336,7 @@ statfs_t *buf;
 	}
 	else
 # endif
-	if (i == 2) {
+	if (i >= 2) {
 		reg.x.ax = 0x7303;
 		reg.x.cx = sizeof(fsbuf);
 # ifdef	DJGPP
@@ -2304,58 +2384,84 @@ statfs_t *buf;
 	return(0);
 }
 
+/*ARGSUSED*/
+static int NEAR alter_findfirst(path, fdp, dbuf, lbuf)
+CONST char *path;
+u_int *fdp;
+struct dosfind_t *dbuf;
+struct lfnfind_t *lbuf;
+{
+	CONST char *cp;
+	char tmp[MAXPATHLEN];
+	int n, len;
+
+#ifndef	_NOUSELFN
+	if (fdp) n = lfn_findfirst(fdp, path, SEARCHATTRS, lbuf);
+	else
+#endif
+	n = dos_findfirst(path, SEARCHATTRS, dbuf);
+	if (n >= 0) return(n);
+	if (errno != ENOENT) return(-1);
+
+	if ((cp = strrdelim(path, 1))) cp++;
+	else cp = path;
+
+	if (isdotdir(cp) == 1) {
+		len = cp - path;
+		memcpy(tmp, path, len);
+		strcpy(&(tmp[len]), curpath);
+#ifndef	_NOUSELFN
+		if (fdp) n = lfn_findfirst(fdp, tmp, SEARCHATTRS, lbuf);
+		else
+#endif
+		n = dos_findfirst(tmp, SEARCHATTRS, dbuf);
+	}
+#ifdef	DOUBLESLASH
+	else {
+		if (isdslash(path)) return(-1);
+		else if (_dospath(path)) /*EMPTY*/;
+		else if (getcurdrv() == '_') return(-1);
+#endif
+		n = dos_findfirst(path, DS_IFLABEL, dbuf);
+	}
+
+	return(n);
+}
+
 int unixstat(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 #ifndef	_NOUSELFN
+# ifndef	_NODOSDRIVE
+	char buf[MAXPATHLEN];
+# endif
 	struct lfnfind_t lbuf;
 	u_int fd;
-#endif
+#endif	/* !_NOUSELFN */
 	struct dosfind_t dbuf;
 	int i;
 
-#ifdef	_NOUSELFN
-	if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-		if (errno && errno != ENOENT) return(-1);
-		if (isdotdir(path) == 1)
-			i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
-		else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
+	if (strpbrk(path, "*?")) {
+		errno = ENOENT;
+		return(-1);
 	}
+#ifdef	_NOUSELFN
+	i = alter_findfirst(path, NULL, &dbuf, NULL);
 #else	/* !_NOUSELFN */
+	fd = (u_int)-1;
 	i = supportLFN(path);
 # ifndef	_NODOSDRIVE
 	if (i < 0 && i > -3) {
-		char buf[MAXPATHLEN];
-
 		dependdosfunc = 0;
 		if (dosstat(regpath(path, buf), stp) >= 0) return(0);
 		else if (!dependdosfunc) return(-1);
 	}
-# endif	/* !_NODOSDRIVE */
-	if (i <= 0) {
-		fd = (u_int)-1;
-		if ((i = dos_findfirst(path, SEARCHATTRS, &dbuf)) < 0) {
-			if (errno && errno != ENOENT) return(-1);
-			if (isdotdir(path) == 1)
-				i = dos_findfirst(curpath, SEARCHATTRS, &dbuf);
-			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
-		}
-	}
-	else {
-		if ((i = lfn_findfirst(&fd, path, SEARCHATTRS, &lbuf)) < 0) {
-			if (errno && errno != ENOENT) return(-1);
-			if (isdotdir(path) == 1)
-				i = lfn_findfirst(&fd, curpath,
-					SEARCHATTRS, &lbuf);
-			else i = dos_findfirst(path, DS_IFLABEL, &dbuf);
-		}
-	}
+# endif
+	if (i <= 0) i = alter_findfirst(path, NULL, &dbuf, &lbuf);
+	else i = alter_findfirst(path, &fd, &dbuf, &lbuf);
 #endif	/* !_NOUSELFN */
-	if (i < 0) {
-		if (!errno) errno = ENOENT;
-		return(-1);
-	}
+	if (i < 0) return(-1);
 
 #ifndef	_NOUSELFN
 	if (fd != (u_int)-1) {
@@ -2380,7 +2486,7 @@ struct stat *stp;
 }
 
 int unixchmod(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
 	struct SREGS sreg;
@@ -2420,8 +2526,8 @@ int mode;
 #ifndef	_NOUSELFN
 #ifdef	USEUTIME
 int unixutime(path, times)
-char *path;
-struct utimbuf *times;
+CONST char *path;
+CONST struct utimbuf *times;
 {
 	time_t t;
 	__dpmi_regs reg;
@@ -2441,8 +2547,8 @@ struct utimbuf *times;
 	}
 #else	/* !USEUTIME */
 int unixutimes(path, tvp)
-char *path;
-struct timeval tvp[2];
+CONST char *path;
+CONST struct timeval *tvp;
 {
 	time_t t;
 	__dpmi_regs reg;
@@ -2486,7 +2592,7 @@ struct timeval tvp[2];
 }
 
 int unixopen(path, flags, mode)
-char *path;
+CONST char *path;
 int flags, mode;
 {
 	struct SREGS sreg;
@@ -2538,10 +2644,10 @@ int flags, mode;
 }
 
 FILE *unixfopen(path, type)
-char *path, *type;
+CONST char *path, *type;
 {
 	FILE *fp;
-	int i, flags;
+	int i, fd, flags;
 
 	path = duplpath(path);
 	i = supportLFN(path);
@@ -2580,8 +2686,8 @@ char *path, *type;
 			break;
 	}
 
-	if ((i = unixopen(path, flags, 0666)) < 0) return(NULL);
+	if ((fd = unixopen(path, flags, 0666)) < 0) return(NULL);
 
-	return(fdopen(i, type));
+	return(fdopen(fd, type));
 }
 #endif	/* !_NOUSELFN */
